# Headers for an EPW file.
EPW_HEADERS = [
    'Year', 'Month', 'Day', 'Hour', 'Minute', 'gibberish', 'Dry-bulb (C)',
    'Dew-point (C)', 'rh (%)', 'Pressure (Pa)', 'ETR (W/m^2)',
    'ETRN (W/m^2)', 'GHInfared (W/m^2)', 'GHI (W/m^2)', 'DNI (W/m^2)',
    'DHI (W/m^2)', 'GH illum (lx)', 'DN illum (lx)', 'DH illum (lx)',
    'Zenith lum (cd/m^2)', 'Wdir (degrees)', 'Wspd (m/s)', 'TotCld (tenths)',
    'OpqCld (tenths)', 'Hvis (m)', 'CeilHgt (m)',
]
# Headers that are actually read from EPW files.
EPW_INTEREST = [
    'GHI (W/m^2)', 'DNI (W/m^2)', 'DHI (W/m^2)', 'Dry-bulb (C)',
    'Dew-point (C)', 'rh (%)', 'Pressure (Pa)', 'Wdir (degrees)',
    'Wspd (m/s)', 'TotCld (tenths)', 'OpqCld (tenths)',
]
# Headers specifying timestamp for EPW files.
EPW_TIMESTAMP = [
    'Year', 'Month', 'Day', 'Hour', 'Minute',
]
# Headers that are actually read from CSV files.
CSV_INTEREST = [
    'GHI (W/m^2)', 'DNI (W/m^2)', 'DHI (W/m^2)', 'Dry-bulb (C)',
    'Dew-point (C)', 'rh (%)', 'Pressure (mbar)', 'Wdir (degrees)',
    'Wspd (m/s)',
]
# Headers specifying timestamp for CSV files.
CSV_TIMESTAMP = [
    'Date (MM/DD/YYYY)', 'Time (HH:MM)',
]
# Header mappings to WeatherTimeSeriesDB attributes.
WTS_ATTR_MAP = {
    'GHI (W/m^2)': 'glo_horz_irr',
    'DNI (W/m^2)': 'dir_norm_irr',
    'DHI (W/m^2)': 'dif_horz_irr',
    'Dry-bulb (C)': 't_dryb',
    'Dew-point (C)': 't_dewp',
    'rh (%)': 'rh',
    'Pressure (Pa)': 'pres_pa',
    'Wdir (degrees)': 'wdir',
    'Wspd (m/s)': 'wspd',
    'TotCld (tenths)': 'tot_cld',
    'OpqCld (tenths)': 'opq_cld',
}



from eagers.building.process.zone_loads import zone_loads
from eagers.building.process.load_sched import load_sched

def load_data(building,weather,date,T_zone,w_zone):
    test_data = {}
    weather_now = {}
    test_data['date'] = []
    test_data['internal_gains'] = []
    test_data['weather'] = {}
    for i in weather:
        if i!='timestamp':
            test_data['weather'][i] = []
    dt = (date[1] - date[0]).seconds
    frost = [[0 for i in range(len(building.cases))]] #need to make as initial condition that is passed from previous time step
    for t in range(len(date)):
        for i in weather:
            weather_now[i] = weather[i][t]
            if i!='timestamp':
                test_data['weather'][i].append(weather[i][t])
        schedules = load_sched(building.schedule,building.holidays,[date[t]],None,False)
        loads,gains,occupancy,mixing,infiltration,S,frost_t,T_mains,water_heat = zone_loads(building,date[t],schedules,weather_now,T_zone,w_zone,dt,frost[t])
        frost.append(frost_t)
        test_data['date'].append(date[t])
        test_data['internal_gains'].append(loads['internal_gain'])
    return test_data


"""User-facing configuration variables.

See eagers.ini for more information on configuration variables.
"""

import os

from .text import parse_bool_str


# Environment variables and their defaults.
DEBUG_MODE = parse_bool_str(os.environ.get("EAGERS_DEBUG_MODE", "False"))
USER_DIR = os.environ.get("EAGERS_USER_DIR", "Documents/eagers_user")
USER_DIR_EXCLUDED = os.environ.get(
    "EAGERS_USER_DIR_EXCLUDE_NAME", "IGNORED"
).replace(" ", "").splitlines()



import datetime 
from random import random

from eagers.config.weather_file import WTS_ATTR_MAP
from eagers.basic.hdf5 import DatetimeFloatConverter as DFC


def get_data(data, date, network_names):
    """Load data corresponding to the given dates. Returns a dictionary
    with the following structure:
    {
        'timestamp': 1-D list of datetime.datetime
        'weather': {
            weather fields: 1-D lists of floats
        'Node1': {
            'network': network name (e.g. 'electrical')
            'demand': 1-D list of floats
        },
        'Node2': {...},
        ...
        'Node<n>': {...},
    }

    Positional arguments:
    data - (TestData) Body of data to be indexed.
    date - (list or array-like of datetime) Date vector for which data
        should be obtained.
    network_names - (list of str) Network names for which data should be
        guaranteed. If one of these names is not found in the data,
        zeros will be returned for that network.
    """
    # Initialize return structure.
    select_data = {}
    table_list = []
    header_list = {}
    d_type = {}
    d_node = {}
    for net in network_names:
        if net in data['nodedata_network_info']:
            for n in range(len(data['nodedata_network_info'][net])):
                cat = list(data['nodedata_network_info'][net][n].keys())
                cat.remove('network')
                cat.remove('node')
                for c in cat:
                    if c.find('history')<0:
                        t_name = data['nodedata_network_info'][net][n][c][0]
                        if not t_name in table_list:
                            table_list.append(t_name)
                            header_list[t_name] = []
                            d_type[t_name] = []
                            d_node[t_name] = []
                        header_list[t_name].append(data['nodedata_network_info'][net][n][c][1])
                        d_type[t_name].append(c)
                        d_node[t_name].append(data['nodedata_network_info'][net][n]['node'])
                        if not d_node[t_name][-1] in select_data:
                            select_data[d_node[t_name][-1]] = {}
                            select_data[d_node[t_name][-1]]['network'] = net
                        if not d_type[t_name][-1] in select_data[d_node[t_name][-1]]:
                            select_data[d_node[t_name][-1]][d_type[t_name][-1]] = {}
    select_data['timestamp'] = [d for d in date]
    d_streams = sum([len(header_list[i]) for i in header_list])  # number of streams of data so that noise can be created independant for each   
    ns = len(date)
    for table in table_list:
        x_query = validate_dates(date,data['nodedata'],table)# Query x values for interpolation.
        x_query_sorted = sorted(x_query)
        if len(x_query)>1:
            dts = 2 * (date[1] - date[0]).total_seconds()
            dte = 2 * (date[-1] - date[-2]).total_seconds()
        else:
            dts = 2 * 3600
            dte = 2 * 3600
        t_data_np = data['nodedata'].read_timestamp_range(table, x_query_sorted[0]-dts, x_query_sorted[-1]+dte)
        t_data = dict(timestamp=t_data_np['timestamp'].tolist())
        t_data_dt = DFC.f2d_arr2lst(t_data_np['timestamp'])
        t_data.update({fk: t_data_np[fk].tolist() for fk in header_list[table]})
        x_vals = t_data['timestamp']
        date_float = DFC.d2f_lst2lst(date)
        x_vals = reverse_validate_date_shift(date, t_data_dt)
        interpolated_data = interp_data(x_vals, t_data, date_float, header_list[table], None, x_vals[1]-x_vals[0])
        j = 0
        for header in header_list[table]:
            select_data[d_node[table][j]][d_type[table][j]] = interpolated_data[header]
            j +=1

    select_data['weather'] = get_weather(data['weather'], date)
    # # New method - reads TMY weather data correctly.  Bypasses
    # # validate_dates(), reverse_validate_date_shift(), interp_data():
    # weather_np = data.weather.read_timestamp_range(
    #     data.weather.tablenames[0],
    #     date[0],
    #     date[-1],
    # )
    # select_data["weather"] = {
    #     k: weather_np[k].tolist()
    #     for k in weather_np.dtype.names if k != "timestamp"
    # }

    # Create noise if interpolation is needed and variability is specified.
    variability = 0
    #TODO: Get variability from relevant source.
    if variability > 0:
        noise = create_noise(ns,d_streams,variability)# Create noise vector.
        i = 0
        for table in table_list:
            j = 0
            for header in header_list[table]:
                select_data[d_node[table][j]][d_type[table][j]][header] = [(1+float(noise[i][t]))*select_data[d_node[table][j]][d_type[table][j]][header][t] for t in range(ns)]
                j+=1
                i+=1

    return select_data


def get_weather(data, date):
    table = data.tablenames[0]
    # x query values for interpolation.
    if data.table_designation(table) == 'tmy':
        x_query = DFC.d2f_lst2lst(date)
    else:
        x_query = validate_dates(date,data,table)
    x_query_sorted = sorted(x_query)
    if len(x_query)>1:
        dts = int(2 * (date[1] - date[0]).total_seconds())
        dte = int(2 * (date[-1] - date[-2]).total_seconds())
    else:
        dts = 2 * 3600
        dte = 2 * 3600
    t_data_np = data.read_timestamp_range(table, x_query_sorted[0]-dts, x_query_sorted[-1]+dte)
    t_data = dict(timestamp=t_data_np['timestamp'].tolist())
    t_data_dt = DFC.f2d_arr2lst(t_data_np['timestamp'])
    # Don't need all weather headers unless there is a building model.
    weather_headers = [h for h in WTS_ATTR_MAP.values()
        if h in t_data_np.dtype.names]
    t_data.update({fk: t_data_np[fk].tolist() for fk in weather_headers})
    date_float = DFC.d2f_lst2lst(date)
    x_vals = reverse_validate_date_shift(date, t_data_dt)
    dtq1 = x_vals[1] - x_vals[0]
    weather = interp_data(x_vals, t_data, date_float, weather_headers, None, dtq1)
    return weather


def create_noise(ns,nd,variability):
    z = [[random() for t in range(ns)] for i in range(nd)]
    # Scaled noise to a signal with average magnitude of 1.
    noise = [[] for i in range(nd)]
    for i in range(nd):
        noise[i] = [0]
        noise[i].append((random()- 0.5) * variability)
        if noise[i][1]>noise[i][0]:
            b=1
        else:
            b = -1
        c = noise[i][1]
        for n in range(2, ns):
            # If the noise is increasing the probability is such that the noise
            # should continue to increase, but as the noise approaches the peak
            # noise magnitude the probability of switching increases.
            if (c > 0 and noise[i][n-1] > 0) or (c < 0 and noise[i][n-1] < 0):
                # Constant 2 = 97.7% probability load changes in same direction,
                # 1.5 = 93.3%, 1 = 84.4%, .5 = 69.1%.
                a = 2 * (variability - abs(noise[i][n-1])) / variability
            else:
                a = 2
            if abs(z[i][n]) > 0.68:
                # Only changes value 50% of the time.
                # c is positive if abs(noise) is increasing, negative if
                # decreasing.
                c = b * (z[i][n]+a)
                if c>0:
                    b=1
                elif c<0:
                    b = -1
                noise[i].append(noise[i][n-1] + c*variability)
            else:
                noise[i].append(noise[i][n-1])
    return noise


def validate_dates(pull_date,data,table):
    # Make checks on the timestamps available for demand data.
    date = [d for d in pull_date]
    first_timestamp = DFC.f2d_sgl2sgl(data.read(table,0,1,field='timestamp')[0])
    last_timestamp = DFC.f2d_sgl2sgl(data.read(table,-1,field='timestamp')[0])
    # Move date up to one day past where the TestData timestamps begin,
    # if needed (starting simulation without historical data).
    if date[0] < first_timestamp:
        # Add whole # (rounded up) of days.
        i = 0
        while i<len(date) and date[i] < first_timestamp:
            date[i] += datetime.timedelta(days=1)
            if date[i] >= first_timestamp:
                i += 1
    if date[-1] > last_timestamp:
        # Subtract whole # of days.
        i = len(date) - 1
        while i>=0 and date[i] > last_timestamp:
            date[i] -= datetime.timedelta(days=1)
            if date[i] <= last_timestamp:
                i -= 1
    date = DFC.d2f_lst2lst(date)
    return date


def reverse_validate_date_shift(pull_date, time_data):
    # Move date up to one day past where the TestData timestamps begin,
    # if needed (starting simulation without historical data).
    first_timestamp = min(pull_date)
    last_timestamp = max(pull_date)
    if time_data[0] > first_timestamp:
        # Subtract whole number (rounded up) of days.
        i = 0
        while i < len(time_data) and time_data[i] > first_timestamp:
            if i == 0 or time_data[i] - datetime.timedelta(days=1) > time_data[i-1]:
                time_data[i] -= datetime.timedelta(days=1)
            if time_data[i] <= first_timestamp or (i > 0 and time_data[i] - datetime.timedelta(days=1) <= time_data[i-1]):
                i += 1
    if time_data[-1] < last_timestamp:
        # Add whole number of days until > last pull date, but not
        # exceeding the value after it.
        i = len(time_data) - 1
        while i >= 0 and time_data[i] < last_timestamp:
            if i == len(time_data) - 1 or time_data[i] + datetime.timedelta(days=1) < time_data[i+1]:
                time_data[i] += datetime.timedelta(days=1)
            if time_data[i] >= last_timestamp or (i < len(time_data) - 1 and time_data[i] + datetime.timedelta(days=1) >= time_data[i+1]):
                i -= 1
    time_data = DFC.d2f_lst2lst(time_data)
    return time_data


def interp_data(x_vals, data, x_query, f, n_input, dtq1):
    # If request data interval is >=2x test data interval, average points
    # If request data interval is <1/2 test data interval, create intermediate points with interpolation
    # Otherwise if timedelta of request and data are within 10% use nearest point to avoid smoothing effect of interpolation
    select_data = {}
    for fk in f:
        select_data[fk] = interp_numerical(data[fk], x_query, x_vals, dtq1)
        if isinstance(n_input, (float, int)):
            select_data[fk] = [select_data[fk][t][n_input] for t in range(len(select_data[fk]))]
    return select_data


def interp_numerical(data,query,timestamp,dtq1):
    n_s = len(query)
    if isinstance(data[0], (float,int)):
        nodes = False
        select = [0 for i in range(n_s)]
    elif isinstance(data[0], list):
        nodes = True
        n = len(data[0])
        select = [[0 for j in range(n)] for i in range(n_s)]
    i = 0
    while timestamp[i]<(query[0]-dtq1):
        i+=1
    for t in range(n_s):
        if query[t]>timestamp[i+1] and len(timestamp)>i+2 and query[t]<=timestamp[i+2]:
            i+=1
        if query[t]<=timestamp[i+1]: #interpolating
            r0 = (timestamp[i+1] - query[t])/(timestamp[i+1] - timestamp[i]) 
            if t<n_s-1 and query[t+1]> timestamp[i+1] and r0<=0.1: #data are within 10% use nearest point to avoid smoothing effect of interpolation
                r0 = 0
            elif t>0 and query[t-1]<timestamp[i] and  r0>=0.9:
                r0 = 1
            if nodes:
                select[t] =  [(r0*data[i][j] + (1-r0)*data[i+1][j]) for j in range(n)]
            else:
                select[t] =  (r0*data[i] + (1-r0)*data[i+1])
            if r0==0 and len(timestamp)>i+2:
                i+=1
        else: #summing and averaging
            if nodes:
                d = [0 for j in range(n)]
            else:
                d = 0
            dtt = query[t]-timestamp[i]
            while query[t]>timestamp[i]:
                dt = min(query[t]-timestamp[i],timestamp[i+1]-timestamp[i])
                if nodes:
                    d = [d[j] + dt*data[i+1][j] for j in range(n)]
                else:
                    d += dt*data[i+1]
                i += 1
            if nodes:
                select[t] =  [d[j]/dtt for j in range(n)]
            else:
                select[t] =  d/dtt
    return select



from eagers.basic.get_data import get_data
from eagers.basic.all_demands import all_demands, add_min_building_demand, subtract_renewable
from eagers.basic.constrain_min_max import constrain_min_max
from eagers.solver.test_min_cases import test_min_cases
from eagers.solver.dispatch_step import building_forecast_now
from eagers.simulate.renewable_output import renewable_output
from eagers.forecasting.hydro_forecast import hydro_forecast
from eagers.update.update_matrices_step import update_matrices_step
from eagers.basic.marginal_cost import net_marginal_cost
from eagers.basic.best_eff import best_eff
from eagers.building.process.forecast_dr_capacity import forecast_dr_capacity


def automatic_ic(gen,building,fluid_loop,market, subnet, date, one_step, options, test_data, building_observer, scale_cost):
    """Flag values:
    0 -- Standard operation.
    1 -- No feasible combination.
    """
    data_t0 = get_data(test_data, date, subnet['network_names'])
    if 'weather' in data_t0 and 'dir_norm_irr' in data_t0['weather']:
        data_t0['renewable'] = [renewable_output(gen[i], date, data_t0['weather']['dir_norm_irr']) \
                                if gen[i]['type'] in ['Renewable','Solar','Wind'] and gen[i]['enabled'] \
                                else [] for i in range(len(gen))]
    data_t0['building'],_,_,_ = forecast_dr_capacity(building,building_observer,data_t0['weather'],date)
    data_t0 = hydro_forecast(data_t0,test_data,date,subnet,[],options)

    demand = all_demands(data_t0,subnet,[0])
    ic_observer = load_ic_observer(gen,fluid_loop,one_step,data_t0)

    # Update marginal cost.
    net_abbrev = [subnet[net]['abbreviation'] for net in subnet['network_names']]
    marginal = net_marginal_cost(gen, net_abbrev, market, None, scale_cost, None, False)
    c_red = best_eff(gen,scale_cost,marginal)  
    sc = [[j[0]] if len(j)>0 else [] for j in scale_cost]
    flt = ic_observer['fluid_loop_temperature']
    flc = ic_observer['fluid_loop_capacitance']
    bat = ic_observer['building_avg_temp']
    fb = building_forecast_now(data_t0['building'],0)
    renewable = {}
    for j in range(len(gen)):
        if gen[j]['type'] in ['Solar','Wind']:
            renewable[gen[j]['name']] = data_t0['renewable'][j][0]
    min_power, max_power = constrain_min_max(gen, 'unconstrained', None, options['resolution'], 0)
    update_matrices_step(gen, market, subnet, options, one_step, fb, renewable, demand,
        sc, marginal, None, options['resolution'], min_power, max_power, None, flt, flc, bat)
    modify_hydro_limits(one_step,subnet,gen,marginal,data_t0)
    add_min_building_demand(data_t0,subnet,demand,[0])
    subtract_renewable(renewable,subnet,demand)
    _, cost, _, disp_comb,_,_,_,_ = test_min_cases(one_step, gen, options, demand, c_red, None,None,None)
    if len(disp_comb) > 0:
        flag = 0
        i_best = cost.index(min(cost))
        ic = disp_comb[i_best]
    else:
        ic = []
        flag = 1
    return ic,data_t0,flag

def load_ic_observer(gen,fluid_loop,one_step,data_t0):
    ic_observer = {}
    n_g = len(gen)
    n_l = len(one_step['organize']['transmission'])
    if not data_t0['building'] is None:
        n_b = len(data_t0['building']['T_avg'][0])
        ic_observer['building_avg_temp'] = [data_t0['building']['T_avg'][0][b] for b in range(n_b)]
    else:
        n_b = 0
        ic_observer['building_avg_temp'] = []
    n_fl = len(fluid_loop['name'])
    ic_observer['fluid_loop_temperature'] = [0 for l in range(n_fl)]
    ic_observer['fluid_loop_capacitance'] = [0 for l in range(n_fl)]
    
    for i in range(n_fl):
        # Set upper and lower bound equal to this temperature to
        # initialize at steady state.
        state = one_step['indices'][n_g+n_l+n_b+i][0][0]
        one_step['lb'][state] = fluid_loop['nominal_return_temperature'][i] - 0.01
        one_step['ub'][state] = fluid_loop['nominal_return_temperature'][i]
        ic_observer['fluid_loop_temperature'][i] = fluid_loop['nominal_return_temperature'][i]
        ic_observer['fluid_loop_capacitance'][i] = fluid_loop['fluid_capacity'][i]*fluid_loop['fluid_capacitance'][i] #Water capacity in kg and thermal capacitance in kJ/kg*K to get kJ/K
    return ic_observer

def modify_hydro_limits(qp,subnet,gen,marginal,data):
    #set hydro gen to produce steady-state, and put substantial price on variation
    gen_names = [gen[i]['name'] for i in range(len(gen))]
    for n in range(qp['network']['electrical']['nodes']):
        equip = subnet['electrical']['equipment'][n]
        req = qp['organize']['balance']['electrical'][n]
        for j in range(len(equip)):
            i = gen_names.index(equip[j])
            if gen[i]['type'] =='HydroStorage':
                states = qp['indices'][i]
                s = states[0]
                outflow = data['hydro']['outflow'][gen[i]['flowcolumn']]
                hydro_power = outflow/gen[i]['stor']['power2flow'] #missing time component?
                #move expected generation to RHS, so that cost penalty can refer to deviations from that power
                qp['beq'][req] -= hydro_power
                qp['lb'][s] -= hydro_power #change in storage for this power output
                qp['ub'][s] -= hydro_power #change in storage for this power output
                #update the equality with NewPower*conversion + spill - Outflow = -nominal PowerGen Flow
                h = qp['organize']['balance']['hydro'].index(n)
                qp['beq'][qp['organize']['hydro_equalities'][h]] = -outflow
                ## assign a cost to deviations
                qp['f'][states[0]] = marginal['e']
                qp['h'][states[0]] = 2*marginal['e']/(.025*hydro_power) #factor of 2 because its solving C = 0.5*x'*H*x + f'*x



import numpy as np

def window_convection(windows,T_int,T_i,cp_air):
    ## correlations for changing tilt angle (page 362)
    ## note temperatures must be in Kelvin!!
    air_density = 1.204 #kg/m^3 %standard air density (1.204 kg/m3) adjusted for the local barometric pressure (standard barometric pressure corrected for altitude, ASHRAE 1997 HOF pg. 6.1).
    g = 9.81 #m/s^2 gravity
    n_win = len(windows['name'])
    deltaT = [T_int[i]-T_i[i] for i in range(n_win)]
    nusselt = [0 for i in range(n_win)]
    h = [0 for i in range(n_win)]
    tilt = []
    for i in range(n_win):
        if abs(windows['normal'][i][2])>0:
            tilt.append(abs(float(np.degrees(np.arctan((windows['normal'][i][0]**2+windows['normal'][i][1]**2)**.5/windows['normal'][i][2])))))
        else:
            tilt.append(90)
    for i in range(n_win):
        if deltaT[i]>0:
            tilt[i] = 180 - tilt[i]
    T_mf = [T_i[i] + 0.25*deltaT[i] for i in range(n_win)]
    mu = [3.723e-6 + 4.94e-8*T_mf[i] for i in range(n_win)]
    lam = [2.873e-3 + 7.76e-5*T_mf[i] for i in range(n_win)]
    Ra_h = [air_density**2*windows['height'][i]**3*g*cp_air[i]*abs(deltaT[i])/(T_mf[i]*mu[i]*lam[i]) for i in range(n_win)]
    Ra_cv = []
    for i in range(n_win):
        if tilt[i]<=179:
            Ra_cv.append(2.5e5*float(np.exp(.72*tilt[i])/np.sin(np.radians(tilt[i])))**(1/5))
        else:
            Ra_cv.append(0)

    for i in range(n_win):
        if tilt[i]<15:
            nusselt[i] = 0.13*Ra_h[i]**(1/3)
        elif tilt[i]>=15 and tilt[i]<=90 and Ra_h[i]<=Ra_cv[i]:
            nusselt[i] = 0.56*(Ra_h[i]*float(np.sin(np.radians(tilt[i]))))**(1/4)
        elif tilt[i]>=15 and tilt[i]<=90 and Ra_h[i]>Ra_cv[i]:
            nusselt[i] = 0.13*(Ra_h[i]**(1/3) - Ra_cv[i]**(1/3)) + 0.56*(Ra_cv[i]*float(np.sin(np.radians(tilt[i]))))**(1/4)
        elif tilt[i]>90 and tilt[i]<=179:
            nusselt[i] = 0.56*(max(min(Ra_h[i],1e11),1e5)*float(np.sin(np.radians(tilt[i]))))**(1/4)
        elif tilt[i]>179:
            nusselt[i] = 0.58*(min(Ra_h[i],1e11))**(1/5)
        h[i] = nusselt[i]*(lam[i]/windows['height'][i])
    return h


from numpy import cos, deg2rad

from eagers.basic.solar_calc import solar_calc
from eagers.setup.update_qpform_all import assign_type


def renewable_output(gen, date, irradiation,location = None):
    """date is the vector of time points at which you request data."""
    panel = None
    turbine = None
    if not isinstance(gen,dict):
        if assign_type(gen) is 'Solar':
            panel = gen.__dict__
            panel['location'] = location
        elif assign_type(gen) is 'Wind':
            turbine = gen.__dict__
    else:
        if gen['type'] == 'Solar':
            panel = gen
        elif gen['type'] ==  'Wind':
            turbine = gen
    power = []
    if not panel is None:
        _, _, azimuth, zenith = solar_calc(panel['location']['longitude'],panel['location']['latitude'],panel['location']['time_zone'],date)
        if panel['tracking'] == 'fixed':
            for t in range(len(azimuth)):
                power.append(panel['size_m2']*(irradiation[t]/1000)*cos(deg2rad(zenith[t]-panel['tilt'])) \
                * max([0, cos(deg2rad(azimuth[t] - panel['azimuth']))])* panel['eff'])
        elif panel['tracking'] in ['1_axis', '1axis']:
            print("Did you mean 'one_axis'?")
        elif panel['tracking'] == 'one_axis':
            for t in range(len(zenith)):
                power.append(panel['size_m2']* (irradiation[t]/1000) * cos(deg2rad(zenith[t] - panel['tilt']))*panel['eff'])
        else:  # Dual axis.
            for t in range(len(irradiation)):
                power.append(panel['size_m2']*(irradiation[t]/1000)*panel['eff'])
    if not turbine is None:
        print('Need to update renewable_output.py for Wind.')
        pass
    return power



from eagers.basic.result_template import result_template
from eagers.basic.get_data import get_data, get_weather
from eagers.update.dispatch_record import  record_initial_conditions
from eagers.building.process.building_warmup import building_warmup, warm_up_date

def initialize_observer(gen,subnet,fluid_loop,all_data_nodes,names, zones, pl,test_data,building_observer,ic,stor_perc,date,data_t0,project_name):
    #operating state of generators, buildings, and fluid loops
    observer = {}
    observer['future'] = result_template(all_data_nodes,names, zones, pl)
    observer['history'] = result_template(all_data_nodes,names, zones, pl)
    n_g = len(names['components'])
    n_l = len(names['lines'])
    observer['timestamp'] = date
    observer['gen_names'] = names['components']
    observer['gen_state'] = [0 for i in range(n_g)]
    observer['gen_status'] = [True for i in range(n_g)]
    observer['stor_state'] = [0 for i in range(n_g)]
    observer['line_names'] = names['lines']
    observer['line_flow'] = [0 for i in range(n_l)]
    lb =  [0 for i in range(n_g)]
    l1 = ['ElectricGenerator','CombinedHeatPower','Chiller','Heater','Electolyzer','HydrogenGenerator','CoolingTower']
    l2 = ['ElectricStorage','ThermalStorage','HydrogenStorage']
    for i in range(n_g):
        observer['gen_state'][i] = ic[i]
        if gen[i]['type'] in l1:
            states = gen[i]['states'][-1]
            for j in range(len(states)):
                lb[i] +=gen[i][states[j]]['lb'][-1]
            if ic[i]<lb[i]:
                observer['gen_status'][i] = False
        elif gen[i]['type'] in l2:
            observer['stor_state'][i] = stor_perc/100 * gen[i]['stor']['usable_size']
        elif gen[i]['type'] == 'HydroStorage':
            observer['stor_state'][i] = gen[i]['start_wy']*gen[i]['stor']['usable_size']
        else:
            observer['gen_state'][i] = ic[i]
    b_keys = list(building_observer.keys())
    for k in b_keys:
        observer[k] = building_observer[k]
    n_fl = len(fluid_loop['name'])
    observer['fluid_loop_name'] = fluid_loop['name']
    observer['fluid_loop_temperature'] = [0 for i in range(n_fl)]
    observer['fluid_loop_capacitance'] = [0 for i in range(n_fl)]
    for i in range(n_fl):
        observer['fluid_loop_temperature'][i] = fluid_loop['nominal_return_temperature'][i]
        observer['fluid_loop_capacitance'][i] = fluid_loop['fluid_capacity'][i]*fluid_loop['fluid_capacitance'][i] #Water capacity in kg and thermal capacitance in kJ/kg*K to get kJ/K
    observer['hydro_wy_timestamp'] = None
    observer['hydro_wy_soc'] = None
    
    dispatch = result_template(all_data_nodes,names, zones, pl)
    record_initial_conditions(data_t0, dispatch, subnet, observer, project_name)
    return observer

def initialize_building_observer(buildings,test_weather,date):
    n_b = len(buildings)
    observer = {}
    observer['building_timestamp'] = [buildings[i].sim_date[0] for i in range(n_b)]
    observer['building_name'] = [buildings[i].name for i in range(n_b)]
    observer['building_avg_temp'] = [0 for i in range(n_b)]
    observer['building_zone_temp'] = [[] for i in range(n_b)]
    observer['building_surf_temp'] = [[] for i in range(n_b)]
    observer['building_humidity'] = [[] for i in range(n_b)]
    observer['building_supply'] = [[] for i in range(n_b)]
    observer['building_return'] = [[] for i in range(n_b)]
    observer['building_conditioned_floor_area'] = [[] for i in range(n_b)]
    observer['building_frost'] = [[] for i in range(n_b)]
    for i in range(n_b):
        building = buildings[i]
        n = len(building.zones['name'])
        rt = building.plant_loop['exit_temperature']
        for j in range(len(building.plant_loop['name'])):
            if building.plant_loop['type'][j] == 'Heating':
                rt[j] -= building.plant_loop['temperature_difference'][j]
            elif building.plant_loop['type'][j] == 'Cooling':
                rt[j] += building.plant_loop['temperature_difference'][j]
        observer['building_supply'][i] =  building.plant_loop['exit_temperature'] #supply 'tank' temperature
        observer['building_return'][i] =  rt #return 'tank' temperature
        hvac = [0 for z in range(n)]
        for zc in list(building.zone_controls.keys()):
            if building.zone_controls[zc]['type'] == 'Thermostat':
                z = building.zones['name'].index(building.zone_controls[zc]['zone'])
                hvac[z] = 1
        observer['building_conditioned_floor_area'][i] = [building.zones['floor_area'][z]*building.zones['multiplier'][z]*hvac[z] for z in range(n)]
        date = warm_up_date(building,date[0])
        weather = get_weather(test_weather,date)
        weather['timestamp'] = [d for d in date]
        T_zone,T_surf,humidity,frost = building_warmup(building,weather,date[0],None)
        observer['building_zone_temp'][i] = T_zone[-1]
        observer['building_surf_temp'][i] =  T_surf[-1]
        observer['building_humidity'][i] =  humidity[-1] #initial humidity of each zone
        observer['building_timestamp'][i] = date[0]
        A = observer['building_conditioned_floor_area'][i]
        observer['building_avg_temp'][i] = sum([observer['building_zone_temp'][i][z]*A[z] for z in range(len(A))])/sum(A) #area_weighted_temperature %% %current average building zone temperature (used as IC in optimization)
        observer['building_frost'][i] = frost[-1]
    if n_b>0:
        zones = [len(x) for x in observer['building_zone_temp']]
        pl = [len(x) for x in observer['building_supply']]
    else:
        zones = [0]
        pl = [0]
    return observer,zones,pl


def generate_bids(gen,initial_disp,date,market):
    # This function generates bids using the predicted dispatch and costs, in
    # combination with forecasting
    n_g = len(gen)
    if len(list(market.keys()))>0:
        n_m = len(market['next_closing_time'])
        bid = [[0 for i in range(len(initial_disp)-1)] for j in range(3)] #Bids have 3 parts - capacity, cost, and timestamp

        SP = get_spot_price(gen,date[0])
        RT = market_forecast(gen,date[1])
        #Assuming 1 market connection - update for more than 1 market
        Pgrid_est = [0 for t in range(len(initial_disp[0])-1)]
        for i in range(n_g):
            if (gen[i]['type']=='Utility' and gen[i]['source'] == 'Electricity') or gen[i]['type'] == 'Market':
                Pgrid_est = [Pgrid_est[t] + initial_disp[i][t+1] for t in range(len(initial_disp[0])-1)]#The estimated external supply needed from market = Pspot + Pmarket
        for i in range(n_m):
            if market['type'] =='Real-Time': #will be generating bids for the Real-time market
                market['m_lambda'][i].append(get_market_lambda(SP,RT[i])) #Compare market price 1 hour ahead, to current spot price to determine lambda
                if market['strategy'][i] =='Price-Taker':
                    for j in range(len(Pgrid_est)):
                        if Pgrid_est[j]>=0: #If buying from market, bid $0 to ensure bid is accepted
                            bid[0][j] = Pgrid_est[j]*market['m_lambda'][i][-1]
                            bid[1][j] = 0
                            bid[2][j] = date[j]
                        else: #If selling to market, bid inf to ensure bid is accepted
                            bid[0][j] = Pgrid_est[j]*market['m_lambda'][i][-1]
                            bid[1][j] = float('inf')
                            bid[2][j] = date[j]
                else:
                    pass
                    #Price-Maker strategy would go here
            elif market['type'][i] =='Day-Ahead':
                pass
            #Add here for Day-Ahead
        market = bid_cap_forecast(bid,market,date[0])

        #send bids to market
        market = submit_bid(bid,market)
    return market

def submit_bid(forecast_bid,market):
    #This function simulates the sending of bids to a connected to market

    bid_hour = market['next_closing_time']
    bid = 0
    _,bid,_ = intersect([round(forecast_bid[t][3],5) for t in range(len(forecast_bid))],round(bid_hour,5)) #Get the time the next bid needs to be submitted, round to the 5th decimal

    #add submitted bid to market struct
    if len(bid)>0:
        market['bid']['capacity'].append(forecast_bid[0][bid])
        market['bid']['price'].append(forecast_bid[1][bid])
        market['bid']['time'].append(forecast_bid[2][bid])
    return market

def get_market_lambda(market_price,spot_price):
    #This function calculates the ratio lambda based off of current market and
    #spot prices

    #place holder lambda until ideal ratio is found
    m_lambda = 3/5 #3/5 from scheduled market, 2/5 from spot market

    return m_lambda

def bid_cap_forecast(bid,market,time):
    #This function updates the forecasted bids, knowing that the current time and
    #the next hour have committed awards

    #Initial disp does not have any awarded capacity for the first 2 hours, so
    #assign times
    if len(market['award']['capacity'])==0:
        market['award']['capacity'].append(0)
        market['award']['capacity'].append(0)
        market['award']['time'].append(time)
        market['award']['time'].append(time+(1/24)) #+1hour in datenum
        market['award']['price'].append(0)
        market['award']['price'].append(0)

    if len(market['award']['capacity'])>0: #If bids have been awarded
        _,award_num,bid_num = intersect(market['award']['time'],bid[2]) #Find indices where award time = time within bid horizon
        for k in range(len(award_num)):
            bid[0][bid_num[k]] = market['award']['capacity'][award_num[k]]
            bid[1][bid_num[k]] = market['award']['price'][award_num[k]]
            bid[2][bid_num[k]] = market['award']['time'][award_num[k]]


    market['forecast']['capacity'] = bid[0]
    market['forecast']['price'] = bid[1]
    market['forecast']['time'] = bid[2]
    return market

def get_spot_price(gen,date):
    #TODO creat this function
    sp = 0.05
    return sp

def market_forecast(gen,date):
    #TODO creat this function
    rt = [0.05 for t in range(len(date))]
    return rt

def intersect(time,bid):
    #TODO creat this function
    a = 1
    award_num = [2]
    bid_num = [3]
    return a,award_num,bid_num


"""Update matrices.

Functions:
update_matrices
update_mat_demand
update_mat_costs
update_mat_build
update_mat_renew
update_mat_hydro
update_mat_sr
update_mat_market
"""


def update_matrices(gen, observer, market, subnet, options, qp, date, scale_cost,
        margin_cost, forecast, ec):
    """gen is the set of generators in the project
    building is the set of buildings in the project
    fluid_loop is the set of fluid loops, e.g. cooling tower water loop
    subnet is the agregated networks
    options are the optimization options
    qp is the optimization problem 
    date is the vector of time steps
    scale_cost is the matrix multiplier of cost for each generator at each time step
    margin_cost is the marginal cost of generation (used in the final value of the energy storage)
    forecast is the forecasted loads
    ec is the end condition for the threshold optimization"""
    
    #copy the default vectors
    qp['h'] = [i for i in qp['nom_vectors']['h']]
    qp['f'] = [i for i in qp['nom_vectors']['f']]
    qp['b'] = [i for i in qp['nom_vectors']['b']]
    qp['b_eq'] = [i for i in qp['nom_vectors']['b_eq']]
    qp['lb'] = [i for i in qp['nom_vectors']['lb']]
    qp['ub'] = [i for i in qp['nom_vectors']['ub']]
    qp['const_cost'] = [i for i in qp['nom_vectors']['const_cost']]

    n_g = len(gen)
    n_b = len(observer['building_avg_temp'])
    dt = [(date[t+1] - date[t]).total_seconds()/3600 for t in range(len(date)-1)]
    n_s = len(dt)
    n_l = len(qp['organize']['transmission'])
    n_fl = len(observer['fluid_loop_temperature'])

    #update initial condition
    for i in range(n_g):
        if 'enabled' in gen[i] and not gen[i]['enabled']:
            qp['organize']['enabled'][i] = False
        ic = None
        if gen[i]['type'] in ['ElectricGenerator','CombinedHeatPower','Chiller','Heater','Electolyzer','HydrogenGenerator','CoolingTower']:
            ic = observer['gen_state'][i]
        elif gen[i]['type'] in ['ElectricStorage','ThermalStorage','HydrogenStorage']:
            ic = observer['stor_state'][i]
        elif gen[i]['type'] == 'HydroStorage':
            ic = observer['stor_state'][i]
        if ic != None: #generators and storage devices
            qp['b_eq'][qp['organize']['ic'][i]] = ic
            qp['ub'][qp['organize']['ic'][i]] = ic+1 #+1 just to help solver find feasible
    for i in range(n_fl):
        qp['b_eq'][qp['organize']['ic'][n_g+n_l+n_b+i]] = observer['fluid_loop_temperature'][i]
        qp['ub'][qp['organize']['ic'][n_g+n_l+n_b+i]] = observer['fluid_loop_temperature'][i]+1 #+1 just to help solver find feasible
    
    update_mat_demand(qp, gen, subnet, forecast, n_s)
    update_mat_costs(qp, gen, margin_cost, scale_cost, ec, dt)
    update_mat_build(qp, observer, forecast['building'], dt)
    if 'renewable' in forecast: # subtract renewable generation 
        update_mat_renew(qp, gen, forecast)
    if 'hydro' in subnet['network_names']:
        update_mat_hydro(qp, gen, observer, subnet, forecast, date)
    if options['spin_reserve']:
        update_mat_sr(qp, options, forecast, dt, n_g)


def update_mat_demand(qp, gen, subnet, forecast, n_s):
    """Update demands and storage self-discharge."""
    g_names = [gen[i]['name'] for i in range(len(gen))]
    for net in subnet['network_names']:
        abbrev = subnet[net]['abbreviation']
        for an in range(len(subnet[net]['nodes'])):#run through all the nodes in this network
            for node in subnet[net]['nodes'][an]: #run through aggregated nodes
                if node in forecast and 'demand' in forecast[node]:# Multiple demands can be at the same node, or none.
                    for t in range(n_s):
                        r_eq = qp['organize']['balance'][net][an][t] #balance at this node
                        qp['b_eq'][r_eq] += forecast[node]['demand'][t]
            for name in subnet[net]['equipment'][an]:  #equipment at this node:
                k = g_names.index(name)
                if 'Storage' in gen[k]['type'] and abbrev in gen[k]['output']:   #if gen[k]['type'] in ['ElectricStorage','ThermalStorage','HydrogenStorage','HydroStorage']:
                    loss = gen[k]['stor']['self_discharge']*gen[k]['stor']['usable_size']*gen[k]['stor']['disch_eff']
                    for t in range(n_s):
                        r_eq = qp['organize']['balance'][net][an][t] #balance at this node
                        qp['b_eq'][r_eq] += loss #account for self-discharge losses


def update_mat_costs(qp, gen, margin_cost, scale_cost, ec, dt):
    """update costs"""
    n_g = len(gen)
    n_s = len(dt)
    for i in range(n_g):
        if not qp['indices'][i][1][0] is None:
            if gen[i]['type'] == 'Utility':
                for t in range(n_s):
                    s = qp['indices'][i][t+1]
                    if len(s)>0:
                        qp['f'][s[0]] *= scale_cost[i][t]*dt[t]
                    if len(s)>1 and gen[i]['sellback_rate'] == -1:
                        qp['f'][s[1]] *= scale_cost[i][t]*dt[t] #sellback is a fixed percent of purchase costs (percent set when building matrices)
                    else:
                        pass #gas utility or constant sellback rate was taken care of when building the matrices
            if gen[i]['type'] in ['ElectricGenerator', 'CombinedHeatPower', 'Chiller', 'Heater', 'HydrogenGenerator', 'CoolingTower', 'Electrolyzer']: #all generators 
                for t in range(n_s):
                    states = qp['indices'][i][t+1]
                    for s in states:
                        qp['h'][s] *= scale_cost[i][t]*dt[t]
                        qp['f'][s] *= scale_cost[i][t]*dt[t]
            if gen[i]['type'] in ['Storage', 'ElectricStorage', 'ThermalStorage','HydrogenStorage','HydroStorage']:
                S = list(gen[i]['output'].keys())
                est_perc_saved_by_storage_over_horizon = 0 #0.01  ## add to storage attributes or generator loading (tunable parameter?)
                val_per_kWh_per_hr = est_perc_saved_by_storage_over_horizon*(margin_cost[S[0]]['max']+margin_cost[S[0]]['min'])/2/sum(dt)
                declining_utility = [1 - t/(len(dt)-1) for t in range(len(dt))]
                if gen[i]['type'] == 'HydroStorage':
                    ss = 1#SOC is second state
                else:
                    ss = 0
                s_end = qp['indices'][i][-1][ss] 
                for t in range(n_s):
                    qp['f'][qp['indices'][i][t+1][ss]] = -val_per_kWh_per_hr*dt[t]*declining_utility[t]
                stor_size = gen[i]['stor']['usable_size']
                if 'u' in gen[i]: #has buffer
                    buff_size = gen[i]['u']['ub']
                else:
                    buff_size = 0    
                if not ec: #full forecast optimization
                    
                    max_value = 1.25*margin_cost[S[0]]['max']
                    min_value = 0.5*margin_cost[S[0]]['min']
                    a1 = -max_value # fitting C = a1*SOC + 0.5*a2*SOC^2 so that dC/dSOC @ 0 = -max & dC/dSOC @ UB = -min
                    a2 = (max_value - min_value)/(stor_size)
                    qp['h'][s_end] = a2 #quadratic final value term loaded into SOC(t=nS)  #quadprog its solving C = 0.5*x'*H*x + f'*x
                    qp['f'][s_end] = a1 #linear final value term loaded into SOC(t=nS)
                    if buff_size>0:
                        for t in range(n_s):
                            su = qp['indices'][i][t+1][-2] #upper buffer state
                            qp['f'][su] = min_value #this is the linear buffer term loaded into the lower & upper buffer
                            qp['h'][su] = (2*max_value-min_value)/buff_size #this is the quadratic buffer term loaded into the lower & upper buffer
                            qp['f'][su+1] = min_value #this is the linear buffer term loaded into the lower & upper buffer
                            qp['h'][su+1] = (2*max_value-min_value)/buff_size #this is the quadratic buffer term loaded into the lower & upper buffer
                else: #used in Online loop when there is a target EC determined by dispatch loop
                    pass
                    #final SOC deviation cost (quadratic cost for error = SOC(end) - EC
                        # if gen[i]['type'] == 'ElectricStorage':
                        #     e_type = 'electrical' 
                        # elif gen[i]['type'] == 'ThermalStorage':
                        #     if gen[i]['source'] == 'heat':
                        #         e_type = 'district_heat' 
                        #     else:
                        #         e_type = 'district_cool' 
                        # elif gen[i]['type'] == 'HydroStorage:
                        #     e_type = 'hydro' 
                        # rows = qp['organize']['inequalities'][i]
                        # n_r = len(qp['indices'][i][1])/n_s
                        # peak_charge_power = gen[i]['ramp']['b'][0]
                        # d_SOC_10_perc = .1*peak_charge_power*(date[-1]-date[0]).seconds/3600  #energy (kWh) if charging at 10#
                        # h[s_end] = -2*min(margin_cost[e_type])/d_SOC_10_perc #quadratic final value term loaded into SOC(t=nS)  #factor of 2 because its solving C = 0.5*x'*H*x + f'*x
                        # qp['f'](s_end) = -min(margin_cost[e_type]) #linear final value term loaded into SOC(t=nS)
                        # n_ic = 0
                        # for j in range(i):
                        #     if len(qp['organize']['ic'])>0:
                        #         n_ic +=1  #order in IC
                        # qp['b_eq'][n_ic] = ic[I]-ec[I]
                        # for t in range(n_s):
                        #     x_n = states[(t-1)*n_t+1] 
                        #     Rn = rows[t*n_r]
                        #     qp['lb'][Xn] = -ec[i] #change lb so that SOC = 0 coresponds to EC
                        #     qp['ub'][Xn] = stor_size - ec[i] #change ub so that SOC = 0 coresponds to EC
                        #     qp['b'][Rn-1] = -buff_size + ec[i] #change lb so that SOC = 0 coresponds to EC (adding EC because there is a -1 in front of SOC in this inequality)
                        #     qp['b'][Rn] = stor_size-buff_size - ec[i] #change lb so that SOC = 0 coresponds to EC          


def update_mat_build(qp, observer, fb, dt,):
    # Within the optimization 5 states are used for a building:
    # #1 represents the air zone temperature setpoint
    # #2 represents the heating required to achieve that zone temperature setpoint
    # #3 represents the cooling required to achieve that zone temperature setpoint
    # #4 represents the temperature in excess of a comfort range (this is penalized)
    # #5 represents the temperature below a comfort range (this is penalized)

    # Building load appears in node equality equation (electric/heating/cooling)
    # 8 inequalities per building:
        #heating: 
            #heating inequality 1 additional heating at lower temp# H>=  UAl*(Th_barl - Ti) + Cap*(Ti - T(i-1))/dt where dt is in seconds
            #heating inequality 2 additional heating at higher temp# H>=  UAi*(Ti-Th_bari) + Cap*(Ti - T(i-1))/dt where dt is in seconds
            #heating inequality 3 minimum heating #  H>=Hmin (done with lower bound)
        #cooling: 
            #Cooling inequality 1 additional cooling at lower temp# C>= UAl*(Tc_barl-Ti) + Cap*(T(i-1)-Ti)/dt where dt is in seconds
            #Cooling inequality 2 additional cooling at higher temp# C>= UAi*(Ti - Tc_bari) + Cap*(T(i-1)-Ti)/dt where dt is in seconds
            #Cooling inequality 3 minimum cooling #  C>=Cmin (done with lower bound)
        #Upper bound inequality
        #Lower bound inequality
    n_s = len(dt)
    n_b = len(observer['building_zone_temp'])
    n_g = len(qp['organize']['dispatchable'])
    n_l = len(qp['organize']['transmission'])
    for i in range(n_b):
        A = observer['building_conditioned_floor_area'][i]
        temperature =  sum([observer['building_zone_temp'][i][z]*A[z] for z in range(len(A))])/sum(A) #area_weighted_temperature %% %current average building zone temperature (used as IC in optimization)
        qp['b_eq'][qp['organize']['ic'][n_g+n_l+i]] = temperature
        qp['ub'][qp['organize']['ic'][n_g+n_l+i]] = temperature+1 #+1 just to help solver find feasible

        for t in range(n_s):
            states = qp['indices'][n_g+n_l+i][t+1]
            #temperature constraint
            qp['lb'][states[0]] = fb['Tmin'][t][i] -1 #maximum deviation beyond comfort bound
            qp['ub'][states[0]] = fb['Tmax'][t][i] +1 #maximum deviation beyond comfort bound

            #electrical energy balance
            r_eq = qp['organize']['building']['electrical_req'][i][t]
            qp['a_eq'][r_eq][states[1]]= -fb['H2E'][t][i] # Electricity = E0 + H2E*Heating + C2E*Cooling
            qp['a_eq'][r_eq][states[2]] = -fb['C2E'][t][i] # Electricity = E0 + H2E*Heating + C2E*Cooling
            qp['b_eq'][r_eq] += fb['E0'][t][i] #Equipment and nominal Fan Power

            r = qp['organize']['building']['r'][i][t]
            #heating inequalities
            if t == 0 and fb['c_min'][t][i] is None: #first state is known (re-optimizing)
                qp['a'][r][states[2]] = -1
                qp['a'][r+1][states[2]] = 1
                qp['b'][r] = -fb['H0'][t][i]
                qp['b'][r+1] = fb['H0'][t][i]+1e-4
            else:
                qp['a'][r][states[0]] = -fb['ua_h'][t][i][0] + fb['Cap'][i]/(3600*dt[t])
                qp['a'][r+1][states[0]] = fb['ua_h'][t][i][1] + fb['Cap'][i]/(3600*dt[t])
                qp['a'][r][states[1]] = -1
                qp['a'][r+1][states[1]] = -1   
                states0 = qp['indices'][n_g+n_l+i][t]      
                qp['a'][r][states0[0]] = -fb['Cap'][i]/(3600*dt[t]) #T(i-1)
                qp['a'][r+1][states0[0]] = -fb['Cap'][i]/(3600*dt[t]) #T(i-1)
                qp['b'][r] = -fb['ua_h'][t][i][0]*fb['th_bar'][t][i][0]
                qp['b'][r+1] = fb['ua_h'][t][i][1]*fb['th_bar'][t][i][1] 
                qp['lb'][states[1]] = fb['h_min'][t][i]               
            
            #cooling inequalities
            if t == 0 and fb['c_min'][t][i] is None: #first state is known (re-optimizing)
                qp['a'][r+2][states[2]] = -1
                qp['a'][r+3][states[2]] = 1
                qp['b'][r+2] = -fb['C0'][t][i]
                qp['b'][r+3] = fb['C0'][t][i]+1e-4
            else:
                qp['a'][r+2][states[0]] = -fb['ua_c'][t][i][0] - fb['Cap'][i]/(3600*dt[t])
                qp['a'][r+3][states[0]] = fb['ua_c'][t][i][1] - fb['Cap'][i]/(3600*dt[t])
                qp['a'][r+2][states[2]] = -1
                qp['a'][r+3][states[2]] = -1
                states0 = qp['indices'][n_g+n_l+i][t]
                qp['a'][r+2][states0[0]] = -fb['Cap'][i]/(3600*dt[t]) #T(i-1)
                qp['a'][r+3][states0[0]] = -fb['Cap'][i]/(3600*dt[t]) #T(i-1)
                qp['b'][r+2] = -fb['ua_c'][t][i][0]*fb['tc_bar'][t][i][0] - fb['Cap'][i]/(3600*dt[t])
                qp['b'][r+3] = fb['ua_c'][t][i][1]*fb['tc_bar'][t][i][1] - fb['Cap'][i]/(3600*dt[t])
                qp['lb'][states[2]] = fb['c_min'][t][i]

            #penalty states (excess and under temperature)
            qp['b'][r+4] = fb['Tmax'][t][i] #upper buffer inequality, the longer the time step the larger the penaly cost on the state is.
            qp['b'][r+5] = -fb['Tmin'][t][i] #lower buffer inequality
            
            #Cost penalty for exceeding temperature bounds
            qp['h'][states[3]] = dt[t]*fb['Discomfort'][t][i]
            qp['h'][states[4]] = dt[t]*fb['Discomfort'][t][i]
            qp['f'][states[3]] = dt[t]*fb['Discomfort'][t][i]
            qp['f'][states[4]] = dt[t]*fb['Discomfort'][t][i]
            
            #Cost penalty for deviation if deadband in nominal temperature window
            qp['b'][r+6] = fb['T_avg'][t][i] #positive deviation inequality
            qp['b'][r+7] = -fb['T_avg'][t][i] #negative deviation inequality
            if fb['deadband'][t][i]:
                qp['f'][states[5]] = dt[t]*fb['Discomfort'][t][i]
            else:
                qp['f'][states[5]] = 0

def update_mat_renew(qp, gen, forecast):
    qp['renewable'] = [[i for i in j] for j in forecast['renewable']]
    n_g = len(gen)
    for i in range(n_g):
        if gen[i]['type'] == 'Solar':
            n_s = len(forecast['renewable'][i]) 
            if 'e' in gen[i]['output']:
                n = gen[i]['subnet_node']['electrical']
                net = 'electrical'
            elif 'dc' in gen[i]['output']:
                n = gen[i]['subnet_node']['direct_current']
                net = 'direct_current'
            for t in range(n_s):
                row = qp['organize']['balance'][net][n][t] #balance at this node (t = 1)
                qp['b_eq'][row] -= forecast['renewable'][i][t]  #put renewable generation into energy balance at correct node
                

def update_mat_hydro(qp, gen, observer, subnet, forecast, date):
    """Updating lower and upper bound for hydro resevoir SOC"""
    qp['organize']['hydro_SOC_offset'] = []
    n_g = len(gen) 
    n_s = len(date)-1 
    for n in range(len(subnet['hydro']['nodes'])): #run through all the nodes in this network
        ## Node inflows (source/sink terms and upstream flow at time t-T ago if t<T)
        #need to be able to forecast sink/source
        for t in range(n_s):
            row = qp['organize']['balance']['hydro'][n][t] #mass balance at this node (t = 1)
            qp['b_eq'][row] -= forecast['hydro']['inflow'][n][t]  #source/sink term (also previous controlled releases that precede date(1))
    for i in range(n_g): #Make sure bounds are between 0 and the maximum usable size for each generator
        if gen[i]['type'] == 'HydroStorage':
            n = gen[i]['subnet_node']['hydro'] #dam #
            if observer['wy_forecast_timestamp'] !=None and date[-1]<observer['wy_forecast_timestamp'][-1]:
                j = 0
                while j<len(observer['wy_forecast_timestamp']) and date[-1]>observer['wy_forecast_timestamp'][j]+1:
                    j+=1
                r = (observer['wy_forecast_timestamp'][j+1] - date[-1])/(observer['wy_forecast_timestamp'][j] - observer['wy_forecast_timestamp'][j])
                qp['organize']['hydro_SOC_offset'][n] = r*observer['hydro_wy_soc'][n][j] +(1-r)*observer['hydro_wy_soc'][n][j+1] # re-center so zero is actual where this target is
                curr_range = max([0.1*gen[i]['stor']['usable_size'], 1+abs(observer['stor_state'][i]-qp['organize']['hydro_SOC_offset'][n])]) 
                for t in range(n_s):
                    s = qp['indices'][i][t+1][1]  #State for Reservoirs
                    qp['ub'][s] = min(curr_range, gen[i]['stor']['usable_size']-qp['organize']['hydro_SOC_offset'][n]) 
                    qp['lb'][s] = max(-curr_range, -qp['organize']['hydro_SOC_offset'][n]) 
            soc_now = observer['stor_state'][i]-qp['organize']['hydro_SOC_offset'][n] 
            qp['b_eq'][qp['organize']['ic'][i]+1] = soc_now #SOC in reservior (2nd ic for dam)
            qp['ub'][qp['organize']['ic'][i]+1] = soc_now+1 #+1 just to help solver find feasible
            qp['lb'][qp['organize']['ic'][i]+1] = soc_now-1 #-1 just to help solver find feasible

    #Adding cost for Line Transfer Penalties to differentiate from spillway flow
    if 'electrical' in subnet and subnet['electrical']['line_names'] !=None:
        for line in range(len(subnet['electrical']['line_names'])):
            if len(qp['indices'][n_g+line][1])>1: #bi-directional transfer with penalty states
                for t in range(n_s):
                    states = qp['indices'][n_g+line][t+1]
                    s_2 = states[1] #penalty from a to b
                    s_3 = states[2] # penalty from b to a
                    qp['f'][s_2] = 0.001  #Adding .1 cent/kWhr cost to every Electric line penalty to differentiate from spill flow
                    qp['f'][s_3] = 0.001  #Adding .1 cent/kWhr cost to every Electric line penalty to differentiate from spill flow
    
    #Adding lower bounds for Hydro lines that have seasonality constraints
    if 'hydro' in subnet and subnet['hydro']['line_names'] !=None:
        for k in range(len(subnet['hydro']['lineNames'])):
            line = subnet['hydro']['lineNumber'][k]
            for t in range(n_s):
                s = qp['indices'][n_g+line][t+1]
                qp['lb'][s] = subnet['hydro']['line_minimum'][k]
                #qp['lb'][s] = subnet['hydro']['line_minimum'][k][date[t+1].month]


def update_mat_sr(qp, options, forecast, dt, n_g):
    n_s = len(dt) 
    c_max = 2 #$2 per kWh at point where shortfall = spin reserve percent*demand or $2 at 5%
    c_min = 0.05 # $0.05 per kWh
    qp['b'][qp['organize']['spin_reserve']] = -forecast['sr_target'] # -shortfall + SRancillary - SR generators - SR storage <= -SR target
    if options['spin_reserve_perc']>5: #more than 5# spinning reserve
        spin_cost = [(c_max*dt[t])/forecast['sr_target'][t] for t in range(n_s)] # -shortfall + SRancillary - SR generators - SR storage <= -SR target
    else:
        spin_cost = []
        for t in range(n_s):
            #TODO (spin reserve on specific nodes)
            r_eq = qp['organize']['balance']['electrical'][0][t]
            spin_cost.append(c_max*dt[t]/(c_min*qp['b_eq'][r_eq])) # -shortfall + SRancillary - SR generators - SR storage <= -SR target

    for t in range(n_s):
        s = qp['organize']['spin_reserve_states'][n_g][t] #cumulative spinning reserve shortfall at t = 1 --> nS
        #     sr_ancillary = qp['organize']['spin_seserve_states'][n_g+1][t] #Ancillary spinning reserve value (negative cost) at t = 1 --> nS
        qp['h'][s] = spin_cost 
        qp['f'][s] = c_min*dt[t]  


def update_mat_market(qp,gen,market,date,dt):
    #Updates the matrices to enforce any current market awards rather than bids
    n_s = len(dt)
    for i in range(len(gen)):
        if gen[i]['type'] =='Market': #update for markets to enforce awards
            if qp['indices'][i] != None:
                if market['include'] == 1: #Update all states if market is included
                    for t in range(n_s):
                        s = qp['indices'][i][t+1][0]
                        qp['ub'][s] = market['forecast']['capacity']
                        qp['lb'][s] = market['forecast']['capacity']
                elif abs(date[0]-market['award_time']).seconds<60: #If there is an award for the current time, but not including forecasted market
                    pass    
                    # award_num =??
                    # if market['award_capacity'][award_num]>=0 #update x if buying
                    #     for t in range(n_s):
                    #         qp.ub(all_states(1:2,1)) = market.award.capacity(award_num:award_num+1,:);
                    #         qp.lb(all_states(1:2,1)) = market.award.capacity(award_num:award_num+1,:);
                    # else: #update Y if selling
                    #     for t in range(n_s):
                    #         qp.ub(all_states(1:2,2)) = market.award.capacity(award_num:award_num+1,:);
                    #         qp.lb(all_states(1:2,2)) = market.award.capacity(award_num:award_num+1,:);



def add_buildings(buildings,subnet,gen):
    """identify what subnet node of electrical heating and cooling the building is on,
    and if there are chillers/heaters connected, disengage the internal HVAC
    """
    nb = len(buildings)
    if nb == 0:
        qpform = []
    else:
        template = {'name':[],'location':[],'electrical_subnet_node':[],'district_heat_subnet_node':[],'heating':[],'district_cool_subnet_node':[],'cooling':[]}
        qpform = [template for i in range(nb)]
        for net in subnet['network_names']:
            if 'buildings' in subnet[net]:
                nodes = subnet[net]['nodes']
                for n in range(len(nodes)):
                    b_num =subnet[net]['buildings'][n]
                    for i in b_num:
                        qpform[i]['name'] = []
                        if net == 'electrical':
                            qpform[i]['electrical_subnet_node'] = n
                            qpform[i]['location'] = subnet['electrical']['location'][n]
                        elif net == 'district_heat':
                            qpform[i]['district_heat_subnet_node'] = n
                            equip = subnet['district_heat']['equipment'][n]
                            for j in range(len(equip)):
                                if gen[j]['type'] in ['heater','chp_generator']:
                                    qpform[i]['heating'] = True
                            if not subnet['district_heat']['connections'][n] is None: #connected to heaters at a different node
                                qpform[i]['heating'] = True
                        elif net == 'district_cool':
                            qpform[i]['district_cool_subnet_node'] = n
                            equip = subnet['district_cool']['equipment'][n]
                            for j in range(len(equip)):
                                if gen[j]['type'] == 'chiller':
                                    qpform[i]['cooling'] = True
                            if not subnet['district_cool']['connections'][n] is None: #connected to chillers at a different node
                                qpform[i]['cooling'] = True
    return qpform


def component_output(gen,subnet,dispatch):
    gen_names = [g['name'] for g in gen]
    source_input = [0 for k in range(len(gen))]
    for i in range(len(gen)):
        if 'eff' in gen[i] and 'cap' in gen[i]:
            k = gen[i]['name']
            source_input[i] = dispatch[k]/eff_interp(gen[i]['cap'],gen[i]['eff'],dispatch[k])
    out = subnet['abbreviation']
    nn = len(subnet['nodes'])
    production = [0 for n in range(nn)]
    for n in range(nn):
        n_equip = subnet['equipment'][n]
        for equip in n_equip:
            k = gen_names.index(equip)
            if out in gen[k]['output']:
                #exceptions for cold-water and combined heat and power. Everything else has 1 input and 1 output
                ##TODO change sign on cw output in chiller and coolingtower
                if out == 'cw' and gen[k]['type'] == 'Chiller':
                    ##TODO add pumping work?
                    production[n] += -dispatch[equip] - source_input[k]
                elif out == 'cw' and gen[k]['type'] == ['CoolingTower']:
                    production[n] += dispatch[equip]
                elif out == 'h' and gen[k]['type'] == 'CombinedHeatPower':
                    production[n] += source_input[k]*eff_interp(gen[k]['cap'],gen[k]['chp_eff'],dispatch[equip])
                elif out == 'e' and gen[k]['type'] == 'ACDCConverter':
                    if dispatch[equip]>0:
                        production[n] += -dispatch[equip]  
                    else:
                        production[n] +=-gen[k]['output']['e'][0][1]*dispatch[equip]
                elif out == 'dc' and gen[k]['type'] == 'ACDCConverter':
                    if dispatch[equip]<0:
                        production[n] += dispatch[equip]
                    else:
                        production[n] += gen[k]['output']['dc'][0][0]*dispatch[equip]
                elif gen[k]['type'] == 'Utility':
                    production[n] += dispatch[equip]
                elif all([j>0 for j in gen[k]['output'][out][-1]]):
                    production[n] += dispatch[equip]
                elif all([j<0 for j in gen[k]['output'][out][-1]]):
                    production[n] += -source_input[k]
    return production

def eff_interp(cap,eff,x):
    if x == 0:
        eff_t = 1 #avoid divide by zero
    elif x<cap[0]:
        eff_t = eff[0]
    else:
        i = 1
        while i<len(cap)-1 and x>cap[i]:
            i+=1
        r = (x-cap[i-1])/(cap[i]-cap[i-1])
        eff_t = (1-r)*eff[i-1] + r*eff[i]
    return eff_t 


import time

from eagers.solver.ecos_qp import ecos_qp
from eagers.solver.sort_solution import sort_solution, sort_gen_disp, sort_eh
from eagers.solver.fix_infeasible import fix_infeasible
from eagers.solver.dispatch_step import dispatch_step
from eagers.solver.cqp_method import cqp_method
from eagers.solver.verify_ramping import verify_ramping
from eagers.update.update_cost import update_cost
from eagers.basic.marginal_cost import net_marginal_cost
from eagers.update.update_matrices import update_matrices
from eagers.update.disable_gen import disable_gen
from eagers.solver.minimize_start_costs import minimize_start_costs
from eagers.basic.all_demands import all_demands, add_min_building_demand

def dispatch_loop(gen,observer,market,subnet,names,op_mat_a,op_mat_b,one_step,options,date,forecast,prediction):
    """
    This is the function that gets called to solve the receding horizon
    dispatch optimization problem. Depending on user selected options, the
    dispatch will be created using mcQP or cQP. This function calculates 
    the optimal dispatch over the forecast horizon.
    
    INPUTS:
    gen        - a list of dictionaries with relevant parameters for each generator
    observer   - a disctionary tracking the current state of generators, buildings and fluid loops and a short history relevant for forecasting methods
    market     - a dictionary with current values for market prices
    subnet     - a dictionary with all of the network nodes, line properties and equipment lists
    op_mat_a   - optimization matrices for fit a
    op_mat_b   - optimization matrices for fit b
    options    - instance of class Options for all user defined optimization options
    date       - list of timestamps for initial condition through last timestep
    forecast   - forecast of demand profiles, weather, and any other parameters
    prediction - estimate taken from previoius solution used to get initial marginal costs

    OUTPUTS:
    solution   - dispatch optimization solution matrix

    FLAGS:
        0 -- Standard operation.
        1 -- Initial dispatch (Fit A) infeasible.
        2 -- Second dispatch (Fit B) infeasible.
    """
    flag = 0# Initialize flag.
    dt = [(date[t+1] - date[t]).total_seconds()/3600 for t in range(len(date)-1)]
    scale_cost = update_cost(date[1:], gen,market)
    net_abbrev = [subnet[net]['abbreviation'] for net in subnet['network_names']]
    margin_cost = net_marginal_cost(gen,net_abbrev,market,prediction, scale_cost, dt, None)
    t_sim = [0 for i in range(3)]# Initialize t_sim
    update_matrices(gen, observer, market, subnet, options, op_mat_a, date,
        scale_cost, margin_cost, forecast, None)

    # STEP 1: Determine initial dispatch.
    t_start = time.time()
    disable_gen(op_mat_a, None)
    x, feasible = ecos_qp(op_mat_a['h'], op_mat_a['f'], op_mat_a['a'], op_mat_a['b'], op_mat_a['a_eq'], op_mat_a['b_eq'], op_mat_a['ub'], op_mat_a['lb'], op_mat_a['x_keep'], op_mat_a['r_keep'], op_mat_a['req_keep'])
    t_sim[0] = time.time() - t_start
    v_h = [[True if j<1 else False for j in s_eh] for s_eh in sort_eh(x, op_mat_a)] #do you value heat generation, or is there tons of extra
    if feasible == 0:
        dispatch1 = [sort_gen_disp(x, op_mat_a, i) for i in range(len(gen))]
        if options['mixed_integer'] and any(op_mat_a['organize']['dispatchable']): # Might be some on/off combinations.
            # STEP 2: dispatch step by step.
            gen_output, stor_power, binary_comb, disp_comb, build_temp, fluid_temp, cost_comb, verified, flag = dispatch_step(
                gen, observer, market, subnet, options, one_step,date, forecast, scale_cost, dt, dispatch1,v_h)
             # #change best dispatch based on re-start costs
            optimal_state,_ = minimize_start_costs(
                one_step ,gen, market, subnet, options, forecast, dispatch1,
                scale_cost, gen_output, stor_power, binary_comb,disp_comb, 
                build_temp, fluid_temp, cost_comb, verified, observer, dt, v_h)
        else:
            optimal_state = dispatch1
        t_sim[1] = time.time() - t_start - t_sim[0]
        # STEP 3: 2nd complete optimization.
        margin_cost = net_marginal_cost(gen,net_abbrev,market,optimal_state, scale_cost, dt, None)
        update_matrices(gen, observer, market, subnet, options, op_mat_b, date,# Update fit B matrices.
            scale_cost, margin_cost, forecast, None)
        if not options['mixed_integer']:
            solution = cqp_method(names, gen, op_mat_b, dispatch1, date)
        else:
            locked = verify_ramping(gen, subnet, op_mat_b, optimal_state, scale_cost, dt)
            disable_gen(op_mat_b, locked)
            x, feasible = ecos_qp(op_mat_b['h'], op_mat_b['f'], op_mat_b['a'], op_mat_b['b'], op_mat_b['a_eq'], op_mat_b['b_eq'], op_mat_b['ub'], op_mat_b['lb'], op_mat_b['x_keep'], op_mat_b['r_keep'], op_mat_b['req_keep'])
            if feasible != 0:
                flag = 2
                print('Dispatch error: Cannot find feasible dispatch.')
                x, feasible = fix_infeasible(op_mat_b)
            v_h = [[True if j<1 else False for j in s_eh] for s_eh in sort_eh(x, op_mat_b)] #do you value heat? record in solution for later
            solution = sort_solution(x, op_mat_b, names, gen, date, v_h)
        t_sim[2] = time.time() - t_start  - sum(t_sim[:2])
    else:
        flag = 1
        x, feasible = fix_infeasible(op_mat_a)
        solution = sort_solution(x, op_mat_a, names, gen, date, v_h)
    solution['timer'] = t_sim
    return solution, flag


"""After-the-fact dispatch result plotting logic from Excel project
file.

Functions:
plot_dispatch_result - Plot a dispatch result.
extract_dispatch_data - Extract dispatch data from HDF5 result file.
"""

import datetime as dt

import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np

from eagers import Project
from eagers.config.network import NETWORK_NAMES
from eagers.config.path_spec import HDF5_SUFFIX, USER_DIR_SIMRESULTS
from eagers.config.plots import (
    DEFAULT_BAR_WIDTH, DEFAULT_FIG_WIDTH, DEFAULT_FIG_HEIGHT,
    ATFDISPATCH_WIDGET_SPACE, SUPTITLE_SPACE)
from eagers.basic.file_handling import find_file_in_userdir
from eagers.basic.hdf5 import h5file_context, DatetimeFloatConverter as DFC


def plot_dispatch_result(project_name, testdata_name):
    """Plot the dispatch result for a simulation that has run all the
    way through.
    """
    # Extract dispatch data from HDF5 file, using component properties
    # from the database to process the results.
    proj = Project.from_excel(project_name, testdata_name)
    disp_data, timestamp, comp_names = extract_dispatch_data(proj)

    # Get demand data.
    demand = proj.test_data.demand.read_timestamp_range(
        proj.options['start_date'],
        proj.options['start_date'] \
            + dt.timedelta(days=proj.options['interval']),
    )
    demand_ts = DFC.f2d_arr2arr(demand['timestamp'])

    # Parameters.
    # Network names should have the order defined in config.
    network_names = [x for x in NETWORK_NAMES
        if x in proj.plant.subnet.network_names]
    n_nets = len(network_names)
    I_MAX = 24  # TODO: Replace this with scrolling ability.
    time_res_sec = int(
        (timestamp[1] - timestamp[0]).astype('m8[s]').astype('u8'))
    bar_width = np.timedelta64(int(DEFAULT_BAR_WIDTH * time_res_sec), 's')

    # Use Matplotlib to plot.
    fig = plt.figure(figsize=(DEFAULT_FIG_WIDTH, DEFAULT_FIG_HEIGHT))
    axs = []
    gs = gridspec.GridSpec(n_nets, 1)
    fig.suptitle(f"Dispatch: {proj.name}", fontsize='x-large')

    # Stacked bar chart of Component generation for each network.
    for i, net in enumerate(network_names):
        axs.append(fig.add_subplot(gs[i]))
        axs[i].set_title(net.title().replace('_', ' '))
        bar_sum = np.zeros(I_MAX)
        for cname in comp_names:
            cdata = disp_data['power'][net][cname][:I_MAX]
            axs[i].bar(
                timestamp[:I_MAX], cdata, bar_width,
                bottom=bar_sum, label=cname)
            bar_sum += cdata

    # Line plot of total demand.
    axs[0].plot(
        demand_ts[:I_MAX], demand['e'][:I_MAX], 'or',
        label='Demand')

    # Add legend.
    axs[0].legend()

    # Auto-format time series x-axis.  This also results in shared tick
    # labels.
    fig.autofmt_xdate()
    # TODO: Align tick marks correctly.
    # https://stackoverflow.com/q/54727603/7232335

    # Constrain layout properly.
    gs.tight_layout(fig,
        rect=[0, ATFDISPATCH_WIDGET_SPACE, 1, 1 - SUPTITLE_SPACE])
    # Remove warning about tight_layout.
    fig.set_tight_layout(False)

    # Add text.
    x_text = 0.1
    y_text = 0.3
    text = fig.text(x_text, y_text, "Some text",
        family='sans-serif', size=14,
        horizontalalignment='left', verticalalignment='top',
        backgroundcolor='#de781f')

    plt.show()


def extract_dispatch_data(project):
    """Extract dispatch data for the given Eagers object.

    Returns a dictionary of the structure:
        data['power'][network][Component name] (kW)
    OR:
        data['soc'][network]['x' (timestamp) or 'y' (value)] (frac)
    """
    # Get data from HDF5 file.
    filepath = find_file_in_userdir(
        USER_DIR_SIMRESULTS, project.name, HDF5_SUFFIX)
    # Open file in read mode.
    with h5file_context(filepath, mode='r') as h5f:
        # Read generator state data from table.
        # Generator name order is assumed to correspond to column order.
        table = h5f.root.result
        # TODO: Get generator names from project file or database.  The
        # generator_names attribute is no longer stored in the HDF5
        # result files.
        comp_names = table.attrs.generator_names
        comp_state = table.col('dispatch/generator_state')
        timestamp = table.col('dispatch/timestamp')

    # Convert array of timestamp floats to array of datetimes.
    # HACK: Add one hour.  Reason for this is unknown.
    timestamp = DFC.f2d_arr2arr(timestamp) + np.timedelta64(1, 'h')

    # Make sense of results using Component properties.
    # Initialize dispatch data structure.
    network_names = [net.name for net in project.plant.network]
    disp_data = dict(
        power={
            net: {comp: np.zeros(len(timestamp)) for comp in comp_names}
            for net in network_names
        },
        soc={net: {'x': [], 'y': []} for net in network_names},
    )

    # Add timestamp information for power data.
    for net in network_names:
        disp_data['power'][net]['timestamp'] = timestamp

    # Populate dispatch data structure.
    for i, comp in enumerate(project.plant.generator):
        # Check the correspondence of Components.
        assert comp.name == comp_names[i], (
            "Component names do not correspond. "
            f"DB: {comp.name}, HDF5: {comp_names[i]}")
        # Use the Component's method for extracting dispatch data.
        for keypath, data in comp.extract_disp_data(
                comp_state[:,i], network_names, timestamp):
            type_, net = keypath.split('/')
            if type_ == 'power':
                disp_data[type_][net][comp_names[i]] = data
            elif type_ == 'soc':
                disp_data[type_][net]['x'].append(timestamp)
                disp_data[type_][net]['y'].append(data)

    return disp_data, timestamp, comp_names



'''
Simulate the initial day X times until initial temperature conditions reach equilibrium
'''

from eagers.building.process.run_building import run_building
from eagers.building.basic.interpolate_weather import interpolate_weather

import datetime
import time

def building_warmup(building,weather,date0,des_day):
    print('Begining building warmup for:  ' + building.name + ' ....')
    t = time.process_time()
    date = warm_up_date(building,date0)
    wu_weather = interpolate_weather(weather,date)
    T_zone = [23 for i in range(len(building.zones['name']))]
    T_surf = [23 for i in range(len(building.ctf['subsurf_names']))]
    humidity = [0.00825 for i in range(len(building.zones['name']))] #initial humidity of each zone
    T_zone, T_surf, humidity, frost,_,_,_,_ = run_building(building,[T_zone],[T_surf],[humidity],wu_weather,date,des_day)
    print('warmup time was:  ' + str(time.process_time()-t) + '  seconds')
    return T_zone, T_surf, humidity, frost

def warm_up_date(building,date0):
    date = []
    wu_days = building.site['min_warm_up'] +.25*(building.site['max_warm_up'] - building.site['min_warm_up']) #warm up days
    if date0>building.sim_date[24*7-1]:
        wu_week = 1
        while wu_days-7*wu_week>0:
            wu_week+=1
        date1 = date0 + datetime.timedelta(hours=-168)
        for x in range(1,int(round(wu_week))):
            date.extend([date1 + datetime.timedelta(hours=i) for i in range(168)])
    else:
        if date0>building.sim_date[23]:
            date1 = date0 + datetime.timedelta(hours=-24)
            for x in range(1,int(round(wu_days))):
                date.extend([date1 + datetime.timedelta(hours=i) for i in range(24)])
        else:
            n = 0
            while date0>building.sim_date[n]:
                n +=1
            for x in range(1,int(round(wu_days))):
                date.extend([date0 + datetime.timedelta(hours=i) for i in range(24-n)])
                date.extend([building.sim_date[0] + datetime.timedelta(hours=i) for i in range(n)])
    return date


from eagers.building.basic.natural_convection import natural_convection
from eagers.building.basic.exterior_convection import exterior_convection

from scipy.sparse import linalg
from scipy.sparse import spdiags

def solve_surface_temp(building,Ts_j0,B,T_surface,T_zone,T,wind_speed,leward,gains):
    
    sig = 5.67e-8 # W/m^2*K^4 Stephan-Boltzman const
    vf = building.zones['view_factors']
    z_index = building.surfaces['zone_index']
    is_absorb = building.ctf['interior_absorb']
    n_sur = len(building.ctf['sur_state1'])
    n_z = len(T_zone)
    n_ext = len(building.ctf['sur_state2'])
    n_ss = len(B)

    T_win = T['windows_int']
    T_interior = [T_surface[building.ctf['sur_state1'][i]] for i in range(n_sur)]
    dT = [T_interior[i] - T_zone[z_index[i]] for i in range(n_sur)]#interior surface convection = h*A*(Tsur - Tzone) 
    h_interior = natural_convection(building.surfaces['normal'],dT,None,building.convection['interior'])
    h_exterior,h_sky = exterior_convection(building,[T_surface[building.ctf['sur_state2'][i]] for i in range(n_ext)],T['air_surface'],wind_speed,leward,T['sky'],building.convection['exterior'])

    A_diag = [i for i in building.ctf['A_diag']]
    A_diag[1] = [A_diag[1][i]+B[i] for i in range(n_ss)]
    
    b = [B[i]*Ts_j0[i] for i in range(n_ss)]    

    ## convective heat transfer with zone
    for i in range(n_sur):
        ss_i = building.ctf['sur_state1'][i]
        z = building.surfaces['zone_index'][i]
        A_diag[1][ss_i] += h_interior[i]
        b[ss_i] += h_interior[i]*T_zone[z]
    ## convective heat transfer with exterior, sky, and ground
    for i in range(n_ext):
        ss_i = building.ctf['sur_state2'][i]
        A_diag[1][ss_i] += h_exterior[i] + h_sky[i]
        b[ss_i] += h_exterior[i]*T['exterior'][i] + h_sky[i]*T['sky']

    ## radiative heat transfer
    for z in range(n_z):
        z_s = len(building.zones['surfaces'][z])
        z_w = len(building.zones['windows'][z])
        for s in range(z_s):
            i = building.zones['surfaces'][z][s]
            ss_i = building.ctf['sur_state1'][i]
            for k in range(z_s):
                j = building.zones['surfaces'][z][k]
                if k>s:
                    if T_interior[i] != T_interior[j]:
                        q_rad_i = sig*vf[i][j]*is_absorb[i]*((T_interior[i]+273)**4 - (T_interior[j]+273)**4)
                        h_rad_i = q_rad_i/(T_interior[i] - T_interior[j])
                        q_rad_j = sig*vf[j][i]*is_absorb[j]*((T_interior[j]+273)**4 - (T_interior[i]+273)**4)
                        h_rad_j = q_rad_j/(T_interior[j] - T_interior[i])
                    else:
                        h_rad_i = sig*vf[i][j]*is_absorb[i]*4*(T_interior[i]+273)**3 #avoid nan by using derivative
                        h_rad_j = sig*vf[j][i]*is_absorb[j]*4*(T_interior[j]+273)**3 #avoid nan by using derivative
                    A_diag[1][ss_i] += h_rad_i
                    b[ss_i] += h_rad_i*T_interior[j]

                    ss_j = building.ctf['sur_state1'][j]
                    A_diag[1][ss_j] += h_rad_j
                    b[ss_j] += h_rad_j*T_interior[i]
            for w in range(z_w):
                j = building.zones['windows'][z][w]
                if T_interior[i] != T_win[j]:
                    q_rad_win = sig*vf[i][j+n_sur]*is_absorb[i]*((T_interior[i]+273)**4 - (T_win[j]+273)**4)
                    h_rad_win = q_rad_win/(T_interior[i] - T_win[j])
                else:
                    h_rad_win = sig*vf[i][j+n_sur]*is_absorb[i]*4*(T_interior[i]+273)**3
                A_diag[1][ss_i] += h_rad_win
                b[ss_i] += h_rad_win*T_win[j]

    for i in range(n_sur):
        b[building.ctf['interior_surface'][i]] += (gains['interior_surface'][i] + gains['window'][i])/building.surfaces['area'][i]
    for i in range(n_ext):
        b[building.ctf['exterior_surface'][i]] +=  gains['exterior_surface'][i]

    A = spdiags(A_diag,[-1,0,1],n_ss,n_ss)
    new_Ts = linalg.spsolve(A,b)
    new_Ts = [float(j) for j in new_Ts]
    max_e_s = max([abs(T_surface[j]-new_Ts[j]) for j in range(n_ss)])
    T_surface = [0.2*T_surface[j]+0.8*new_Ts[j] for j in range(n_ss)]
    dT = [T_interior[i] - T_zone[z_index[i]] for i in range(n_sur)]#interior surface convection = h*A*(Tsur - Tzone) 
    h_interior = natural_convection(building.surfaces['normal'],dT,None,building.convection['interior'])
    return new_Ts,h_interior,max_e_s


from eagers.basic.gen_limit import gen_limit
from eagers.config.simulation import CHP_HEAT_VALUE

def gen_marginal_cost(gen, dispatch, scale_cost, dt):
    ''' Finds the capacity and marginal cost for each generation component 
    in three categories: spinning reserve, non-spinning reserve, and demand response
    These categories can be thought of as a: turning up a generator already on,
    b: turning on an inactive generator, and c: turning down an active generator'''
    n = 4 #breaks remaining generator capacity into n segments
    n_s = len(dt)
    n_g = len(gen)
    include = ['ElectricGenerator','CombinedHeatPower','Heater','Electrolyzer','HydrogenGenerator','CoolingTower','Chiller']
    locked = [[True for j in range(n_s+1)] for i in range(n_g)]
    for i in range(n_g):
        if gen[i]['type'] in include:
            locked[i] = [dispatch[i][t]>0 for t in range(n_s+1)]
    margin_cost = {}
    s = {'e':['CombinedHeatPower', 'ElectricGenerator'],'h':['Heater'], 'c':['Chiller']}
    #TODO add other dispatchable components
    for i in range(n_g):
        if gen[i]['type'] in ['CombinedHeatPower', 'ElectricGenerator', 'HydrogenGenerator'] and 'dc' in gen[i]['output']:
            s['dc'] = ['CombinedHeatPower', 'ElectricGenerator']
            # Marginal costs of 'electricity' and 'direct_current' will be lumped together.
            break
    max_out,_,_,_ = gen_limit(gen,dispatch,locked,dt,list(s.keys()))
    for k in list(s.keys()):
        inc = [False for i in range(n_g)]
        # Include the Component at i if its primary output is the type
        # of energy we're finding the marginal cost of.
        for i in range(n_g):
            if gen[i]['type'] in s[k]:
                inc[i] = True
        if any(inc):
            margin_cost[k] = {}
            margin_cost[k]['capacity'] = {}
            margin_cost[k]['cost'] = {}
            margin_cost[k]['capacity']['spin_reserve'] = [[[0 for j in range(n)] for t in range(n_s)] for i in range(n_g)]
            margin_cost[k]['capacity']['non_spin'] = [[[0 for j in range(n)] for t in range(n_s)] for i in range(n_g)]
            margin_cost[k]['capacity']['demand_response'] = [[[0 for j in range(n)] for t in range(n_s)] for i in range(n_g)]
            margin_cost[k]['cost']['spin_reserve'] = [[[0 for j in range(n)] for t in range(n_s)] for i in range(n_g)]
            margin_cost[k]['cost']['non_spin'] = [[[0 for j in range(n)] for t in range(n_s)] for i in range(n_g)]
            margin_cost[k]['cost']['demand_response'] = [[[0 for j in range(n)] for t in range(n_s)] for i in range(n_g)]
            for i in range(n_g):
                if inc[i]:
                    sr = [max_out[k][i][t+1] - dispatch[i][t+1] for t in range(n_s)]
                    for t in range(n_s):
                        if sr[t] > 1e-4:
                            disp_range = [dispatch[i][t+1] + j*(max_out[k][i][t+1] - dispatch[i][t+1])/n for j in range(n+1)]
                            margin_cost[k]['capacity']['spin_reserve'][i][t] = [sr[t]/n for j in range(n)]
                            mc = [piecewise_tangent(gen[i],disp_range[j],scale_cost[i][t]) for j in range(n+1)]
                            margin_cost[k]['cost']['spin_reserve'][i][t] = [(mc[j+1] + mc[j])/2 for j in range(n)]
                            margin_cost[k]['cost']['non_spin'][i][t]= [float('inf') for j in range(n)]
                        else:
                            states = gen[i]['states'][-1]
                            disp_range = [gen[i]['lb'] + j*(gen[i]['ub'] - gen[i]['lb'])/n for j in range(n+1)]
                            if disp_range[0]>0:
                                disp_range.insert(0,0)
                            lb = disp_range[1]
                            cost  = [gen[i][states[0]]['f'][-1]*scale_cost[i][t]]
                            if 'start_cost' in gen[i] and gen[i]['start_cost']>0:
                                cost[0] += gen[i]['start_cost']/lb
                            if 'const_cost' in gen[i] and  gen[i]['const_cost']>0:
                                cost[0] += gen[i]['const_cost']*scale_cost[i][t]/lb
                            if gen[i]['type'] == 'Chiller' and  'electrical' in gen[i]['const_demand']:
                                cost[0] += gen[i]['const_demand']['electrical']*min([margin_cost['e']['cost']['spin_reserve'][j][t][0] for j in range(n_g)])/lb
                            mc = [piecewise_tangent(gen[i],disp_range[j+1],scale_cost[i][t]) for j in range(len(disp_range)-1)]
                            cost.extend([(mc[j+1] + mc[j])/2 for j in range(len(disp_range)-2)])
                            margin_cost[k]['capacity']['non_spin'][i][t] = [disp_range[x+1] - disp_range[x] for x in range(len(disp_range)-1)]
                            margin_cost[k]['cost']['non_spin'][i][t] = cost
                            margin_cost[k]['cost']['spin_reserve'][i][t] = [float('inf') for j in range(len(disp_range)-1)]
    return margin_cost


def sort_mc(margin_cost, out, t_max, off_gen, dt):
    '''
    Five attributes tracked as sorting marginal cost
    %1 = cumulative marginal energy, 
    %2 = cumulative marginal cost, used for interpolating to see which blocks are used
    %3 = Generator that increases output, 
    %4 = time point at which generator increases output, 
    %5 = amount that generator increases output (kW)
    '''
    mc = {'capacity':[],'marginal_cost':[],'generator':[],'timestep':[],'duration':[]}
    if not out in margin_cost and out=='dc' and 'e' in margin_cost:
        out = 'e'
    elif not  out in margin_cost and out=='e' and 'dc' in margin_cost:
        out = 'dc'
    n = 4 # must match n in marginal_cost
    n_g = len(margin_cost[out]['capacity']['spin_reserve'])

    for i in range(n_g):
        if off_gen == None or not i == off_gen:
            for t in range(t_max): #+1 because t_max is an index
                if margin_cost[out]['capacity']['spin_reserve'][i][t][0]>0:
                    for j in range(n):
                        mc['capacity'].append(margin_cost[out]['capacity']['spin_reserve'][i][t][j])
                        mc['marginal_cost'].append(margin_cost[out]['cost']['spin_reserve'][i][t][j])
                        mc['generator'].append(i)
                        mc['timestep'].append(t)
                        mc['duration'].append(dt[t])
    if len(mc['capacity'])>0:
        ind = sorted(range(len(mc['marginal_cost'])), key=mc['marginal_cost'].__getitem__)
        cum_cost = [mc['marginal_cost'][ind[0]]*mc['capacity'][ind[0]]*mc['duration'][ind[0]]]
        cum_energy = [mc['capacity'][ind[0]]*mc['duration'][ind[0]]]
        for i in range(len(ind)-1):
            j = ind[i+1]
            c = cum_cost[-1] + mc['marginal_cost'][j]*mc['capacity'][j]*mc['duration'][j]
            e = cum_energy[-1] + mc['capacity'][j]*mc['duration'][j]
            cum_cost.extend([c])
            cum_energy.extend([e])
        mc_sorted = {'cumulative_energy':cum_energy,'cumulative_cost':cum_cost,'generator':[mc['generator'][i] for i in ind],'timestep':[mc['timestep'][i] for i in ind],'capacity':[mc['capacity'][i] for i in ind],'marginal_cost':[mc['marginal_cost'][i] for i in ind]}
    else:
        mc_sorted = None
    return mc_sorted


def piecewise_tangent(gen,dispatch,scale_cost):
    states = gen['states'][-1]
    p = 0
    j = 0
    while p<dispatch and j < len(states):
        x = min([gen[states[j]]['ub'][-1],dispatch-p])
        p +=x
        j +=1
    mc = (gen[states[j-1]]['f'][-1] + x*gen[states[j-1]]['h'][-1])*scale_cost
    return mc

def locate_node_gen(gen,subnet):
    gen_names = [gen[i]['name'] for i in range(len(gen))]
    gen_loc = {}
    for net in subnet['network_names']: #make sure 'e' and 'dc', and 'h' before 'c'
        net_abbrev = subnet[net]['abbreviation']
        gen_loc[net_abbrev] = [None for i in range(len(gen))]
        for n in range(len(subnet[net]['nodes'])):
            agregate_node = subnet[net]['nodes'][n][0]
            gen_num = [gen_names.index(e_name) for e_name in subnet[net]['equipment'][n]]
            for gn in gen_num:
                gen_loc[net_abbrev][gn] = agregate_node
    return gen_loc

def marginal_node(net_abbrev,gen_num,gen_loc,marginal):
    m_node = {}
    for f in list(marginal.keys()):
        if not f == net_abbrev:
            connect_node = []
            for j in gen_num:
                if not gen_loc[f][j] is None and not gen_loc[f][j] in connect_node:
                    connect_node.append(gen_loc[f][j])
            if len(connect_node)>0:
                m_node[f] = min([marginal[f][j] for j in connect_node])
    return m_node

def node_marginal(gen,subnet,dispatch,scale_cost,v_h):
    net_abbrev = [subnet[net]['abbreviation'] for net in subnet['network_names']]
    marginal = {}
    gen_loc = locate_node_gen(gen,subnet)
    for net in subnet['network_names']: #make sure 'e' and 'dc', and 'h' before 'c'
        net_abbrev = subnet[net]['abbreviation']
        marginal[net_abbrev] = {}
        for n in range(len(subnet[net]['nodes'])):
            agregate_node = subnet[net]['nodes'][n][0]
            marginal[net_abbrev][agregate_node] = None
            gen_num = [i for i in range(len(gen)) if gen_loc[net_abbrev][i] == agregate_node]
            gen_node = [gen[i] for i in gen_num]
            gen_scale = [scale_cost[i] for i in gen_num]
            gen_disp = None
            if not dispatch is None:
                gen_disp = [dispatch[i] for i in gen_num]
            m_node = marginal_node(net_abbrev,gen_num,gen_loc,marginal)
            marginal[net_abbrev][agregate_node] = equip_marginal(gen_node,net_abbrev,gen_disp,gen_scale,m_node,v_h)
    for net in subnet['network_names']:
        net_abbrev = subnet[net]['abbreviation']
        mar_net = [marginal[net_abbrev][an] for an in list(marginal[net_abbrev].keys()) if not marginal[net_abbrev][an] is None]
        max_marginal = [max([mar_net[i][t] for i in range(len(mar_net))]) for t in range(len(mar_net[0]))] #vector of maximum marginal cost anywhere on network at each time step 

        for an in marginal[net_abbrev]:
            if marginal[net_abbrev][an] is None:
                marginal[net_abbrev][an] = [2*j for j in max_marginal]
    return marginal

def equip_marginal(gen,net_abbrev,dispatch,scale_cost,source_marginal,v_h):
    #equipment marginal output cost at a particular node 
    n_g = len(gen)
    if n_g>0:
        n_s = max([len(scale_cost[i]) for i in range(len(scale_cost))])
    chp = [i for i in range(n_g) if gen[i]['type'] == 'CombinedHeatPower']
    e_marginal = [None for i in range(n_g)]
    e_min = [None for i in range(n_g)]
    e_max = [None for i in range(n_g)]
    _, ac_dc_eff,conv = check_acdc(gen,net_abbrev)
    for i in range(n_g):
        include = False
        sc = [k for k in scale_cost[i]]
        if net_abbrev in ('e','dc') and net_abbrev in gen[i]['output'] and gen[i]['enabled'] and gen[i]['type'] in ('CombinedHeatPower', 'ElectricGenerator', 'HydrogenGenerator'):
           include = True 
           if v_h and i in chp:
                sc = [k*(1-CHP_HEAT_VALUE) for k in sc]
        elif net_abbrev in ('h') and gen[i]['enabled'] and gen[i]['type'] in ('CombinedHeatPower', 'Heater'):
            include = True 
            if i in chp:
                sc = [k*CHP_HEAT_VALUE for k in sc]
                if not v_h:
                    sc = [k/4 for k in sc]
        elif net_abbrev in ('c') and gen[i]['enabled'] and gen[i]['type'] == 'Chiller': # Cooling marginal cost. # Chillers are unique because the cost in QP.f is zero, so must have heating or electricity done first.
            include = True 
            sc = None
        elif net_abbrev in ('cw') and gen[i]['enabled'] and gen[i]['type'] == 'CoolingTower': # Cooling marginal cost. # Chillers are unique because the cost in QP.f is zero, so must have heating or electricity done first.
            include = True 
            sc = None
        ##TODO complete water and hydrogen
        elif net_abbrev in ('w') and gen[i]['enabled'] and gen[i]['type'] == 'HydroStorage':
            pass
        elif net_abbrev in ('hy') and gen[i]['enabled'] and gen[i]['type'] == 'Electrolyzer':
            pass
        if include and sc is None: #CHILLER or Cooling Tower
            f = list(gen[i]['output'].keys())
            for s in f:
                if s in source_marginal and all([i<=0 for i in gen[i]['output'][s][-1]]):
                    if not dispatch is None:
                        seg = active_seg(gen[i],dispatch[i])
                    else:
                        seg = [0 for t in range(n_s)]
                    e_marginal[i] = [source_marginal[s][t]*(-gen[i]['output'][s][-1][seg[t]]) for t in range(n_s)]
                    e_min[i] = [-gen[i]['output'][s][-1][0]*j for j in source_marginal[s]]
                    e_max[i] = [-gen[i]['output'][s][-1][-1]*j for j in source_marginal[s]]
        elif gen[i]['type'] == 'Utility' and net_abbrev in gen[i]['output']:
            e_marginal[i] = [gen[i]['x']['f']*j for j in sc]
            e_min[i] = [0.5*j for j in e_marginal[i]]
            e_max[i] = [1.5*j for j in e_marginal[i]]
        elif include or i in conv:
            states = gen[i]['states'][-1]
            if not dispatch is None:
                e_marginal[i] = [m_cost(gen[i], dispatch[i][j], sc[j], source_marginal) for j in range(len(sc))]
            else:
                e_marginal[i] = [m_cost(gen[i], None, sc[j], source_marginal) for j in range(len(sc))]
            e_min[i] = [gen[i][states[0]]['f'][-1]*j for j in sc]
            l_state = gen[i][states[-1]]
            e_max[i] = [(l_state['f'][-1]+l_state['ub'][-1]*l_state['h'][-1])*j for j in sc]
            if i in conv:
                e_marginal[i] = [j/ac_dc_eff[i] for j in e_marginal[i]]
                e_min[i] = [j/ac_dc_eff[i] for j in e_min[i]]
                e_max[i] = [j/ac_dc_eff[i] for j in e_max[i]]
    # return e_marginal, e_min, e_max
    min_marginal = None
    if any([True for j in e_marginal if not j is None ]):
        min_marginal = [min([j[t] for j in e_marginal if not j is None]) for t in range(n_s)]
    return min_marginal

def check_acdc(gen,net_abbrev):
    need_ac_dc = False
    ac_dc_eff = [1 for i in range(len(gen))]
    ac_dc = None
    conv = []
    if net_abbrev == 'e' or net_abbrev == 'dc':
        for gen_i in gen:
            if ('e' in gen_i['output'] and net_abbrev == 'dc') or ('dc' in gen_i['output'] and net_abbrev == 'e'):
                need_ac_dc = True
                break
        if need_ac_dc:
            for gen_i in gen:
                if gen_i['type'] == 'ACDCConverter':
                    ac_dc = {}
                    ac_dc['dc_to_ac_eff'] = gen_i['output']['e'][0][1]
                    ac_dc['ac_to_dc_eff'] = gen_i['output']['dc'][0][0]
            if not ac_dc is None:
                for i in range(len(gen)):
                    if gen[i]['enabled'] and gen[i]['type'] in ('CombinedHeatPower', 'ElectricGenerator', 'HydrogenGenerator'):
                        if net_abbrev == 'e' and 'dc' in gen[i]['output']:
                            ac_dc_eff[i] = 1/ac_dc['dc_to_ac_eff']
                            conv.append(i)
                        if net_abbrev == 'dc' and 'e' in gen[i]['output'] and ac_dc:
                            ac_dc_eff[i] = 1/ac_dc['ac_to_dc_eff']
                            conv.append(i)
    return ac_dc, ac_dc_eff, conv

def net_marginal_cost(gen,net_abbrev,market,dispatch, scale_cost, dt, v_h):
    """Calculate the marginal cost of energy for each energy network"""
    n_g = len(gen)
    s = []
    ac_dc = None
    
    for gen_i in gen:
        if isinstance(gen_i['output'],dict):
            f = list(gen_i['output'].keys())
            for k in f:
                if k not in s and k in net_abbrev:
                    s.append(k)
        if gen_i['type'] == 'ACDCConverter':
            ac_dc = {}
            ac_dc['dc_to_ac_eff'] = gen_i['output']['e'][0][1]
            ac_dc['ac_to_dc_eff'] = gen_i['output']['dc'][0][0]

    marginal = {}
    chp = [i for i in range(n_g) if gen[i]['type'] == 'CombinedHeatPower']
    
    if len(chp)>0 and v_h==None:
        v_h = True
    stor_include = {}
    for out in s:
        stor_include[out] = []
    for i in range(n_g):
        if 'stor' in gen[i]:
            f = list(gen[i]['output'].keys())
            stor_include[f[0]].append(i)
    scale = [[i for i in j] for j in scale_cost]
    if 'e' in s or 'dc' in s:
        # Electricity or direct current marginal cost.
        include = []
        dc_to_ac = [1] * n_g
        ac_to_dc = [1] * n_g
        for i in range(n_g):
            if gen[i]['enabled'] and gen[i]['type'] in ('CombinedHeatPower', 'ElectricGenerator', 'HydrogenGenerator'):
                include.append(i)
                if 'dc' in gen_i['output'] and ac_dc:
                    dc_to_ac[i] = 1/ac_dc['dc_to_ac_eff']
                if 'e' in gen_i['output'] and ac_dc:
                    ac_to_dc[i] = 1/ac_dc['ac_to_dc_eff']
        if ac_dc and 'e' in s and 'dc' in s:
            for i in stor_include['dc']:
                dc_to_ac[i] = 1/ac_dc['dc_to_ac_eff']
            for i in stor_include['e']:
                ac_to_dc[i] = 1/ac_dc['ac_to_dc_eff']
            sie_temp = [i for i in stor_include['e']]
            sie_temp.extend(stor_include['dc'])
            stor_include['dc'].extend(stor_include['e'])
            stor_include['e'] = [i for i in sie_temp]
        if v_h:
            for j in chp:
                scale[j] = [k*(1-CHP_HEAT_VALUE) for k in scale[j]]
           
        def adjust_scale(scale,eff):
            sc = []
            for i in range(len(scale)):
                if len(scale[i])==1:
                    sc.append(scale[i][0]*eff[i])
                elif len(scale[i])>1:
                    sc.append([k*eff[i] for k in scale[i]])
                else:
                    sc.append([])
            return sc
        def one_d_list(scale):
            sc = []
            for k in scale:
                if len(k)>0:
                    sc.append(k[0])
                else:
                    sc.append([])
            return sc
        if 'e' in s:
            sc = adjust_scale(scale,dc_to_ac)
            if not dt is None and (dispatch is None or isinstance(dispatch[0],(list))):
                marginal['e'] =  min_max_cost(gen,include,stor_include['e'],sc,dispatch,'e',[],dt)
            else:
                marginal['e'] = m_loop(gen,include,sc,dispatch,'e',None)
        if 'dc' in s:
            sc = adjust_scale(scale,ac_to_dc)
            if not dt is None and (dispatch is None or isinstance(dispatch[0],(list))):
                marginal['dc'] =  min_max_cost(gen,include,stor_include['dc'],sc,dispatch,'dc',[],dt)
            else:
                marginal['dc'] = m_loop(gen,include,sc,dispatch,'dc',None)

    if 'h' in s:
        # Heating marginal cost.
        include = []
        for i in range(n_g):
            if gen[i]['enabled'] and gen[i]['type'] in ['CombinedHeatPower', 'Heater']:
                include.append(i)
        scale = [[i for i in j] for j in scale_cost]
        for j in chp:
            scale[j] = [k*CHP_HEAT_VALUE for k in scale[j]]
            if not v_h:
                scale[j] = [k/4 for k in scale[j]]
        if not dt is None and (dispatch is None or isinstance(dispatch[0],(list))):
            marginal['h'] = min_max_cost(gen,include,stor_include['h'],scale,dispatch,'h',[],dt)
        else:
            marginal['h'] = m_loop(gen,include,one_d_list(scale),dispatch,'h',None)

    if 'c' in s:
        # Cooling marginal cost.
        # Chillers are unique because the cost in QP.f is zero.
        include = []
        for i in range(n_g):
            if gen[i]['enabled'] and gen[i]['type'] == 'Chiller':
                include.append(i)
        if not dt is None and (dispatch is None or isinstance(dispatch[0],(list))):
            marginal['c'] = min_max_cost(gen,include,stor_include['c'],scale_cost,dispatch,'c',marginal,dt)
        else:
            marginal['c'] = m_loop(gen,include,one_d_list(scale_cost),dispatch,'c',marginal)

    if 'w' in s:
        if not dt is None and (dispatch is None or isinstance(dispatch[0],(list))):
            marginal['w'].min = 0
            marginal['w'].max = 1
        else:
            marginal['w'] = 1

    if 'hy' in s:
        marginal['hy'] = marginal['e']

    return marginal


def m_loop(gen, include, scale, step_profile, s, marginal):
    if not include:
        # No generators to be considered in marginal cost.  Return zero.
        return 0
    margin_cost = []
    n_g = len(gen)
    for i in include:
        if step_profile==None:
            sp = None
        else:
            sp = step_profile[i]
        margin_cost.append(m_cost(gen[i],sp,scale[i],marginal))
    m_c = float('nan')
    if not step_profile is None:
        for j in range(len(include)):
            i = include[j]
            if gen[i]['type'] in ['ElectricGenerator', 'CombinedHeatPower', 'Chiller', 'Heater', 'CoolingTower', 'Electrolyzer', 'HydrogenGenerator']:
                lb = gen[i][gen[i]['states'][1][0]]['lb'][-1]
            else:
                lb = gen[i][gen[i]['states'][0]]['lb'] #on or above lower bound
            if step_profile[i]>lb: #on or above lower bound
                if m_c != m_c:
                    m_c = margin_cost[j]
                else:
                    m_c = max([margin_cost[j],m_c])
    if step_profile==None or m_c != m_c: # or m_c==0
        m_c = min(margin_cost)
    for i in range(n_g):
        if gen[i]['type'] == 'Utility' and s in gen[i]['output']:
            if  gen[i]['x']['f']*scale[i]<m_c:
                m_c = gen[i]['x']['f']*scale[i]
    return m_c

def m_cost(gen, set_point, scale, marginal):
    #find the slope of the currently active segment (first segment if no dispatch yet)
    #if it's cost is the output of another kind, e.g. electric chiller, use the electric marginal cost
    states = gen['states'][1]
    seg = active_seg(gen,[set_point])
    m_c = gen[states[seg[0]]]['f'][-1]*scale
    if not set_point is None:
        m_c += (set_point*gen[states[seg[0]]]['h'][-1])*scale
    if m_c == 0:
        for s in list(gen['output'].keys()):
            if all([i<=0 for i in gen['output'][s][-1]]):
                m_c = marginal[s]*(-gen['output'][s][-1][seg[0]])
    m_c = nonneg_marginal(m_c,gen,scale)
    return m_c 

def active_seg(gen,dispatch):
    #find the currently active segment (first segment if no dispatch yet)
    if not dispatch[0] is None:
        states = gen['states'][1]
        seg = [0 for t in range(len(dispatch))]
        for t in range(len(dispatch)):
            set_point = dispatch[t]*1.0
            while seg[t]<len(states)-1 and gen[states[seg[t]]]['ub'][-1]<=set_point:
                set_point -= gen[states[seg[t]]]['ub'][-1]
                seg[t] += 1
    else:
        seg = [0]
    return seg

def nonneg_marginal(m_c,gen,scale):
    if m_c<0: 
        states = gen['states'][1]
        j = 0
        while j<len(states) and m_c<0:
            j = j+1
            m_c = gen[states[j]]['f'][-1]*scale
        if m_c<0:
            m_c = 1e-3
    return m_c

def min_max_cost(gen, include, stor_include, scale, dispatch,s,marginal,dt):
    n_g = len(gen)
    n_s = len(dt)
    min_c = [[0 for t in range(n_s)] for i in range(max([1,len(include)]))]
    max_c = [[0 for t in range(n_s)] for i in range(max([1,len(include)]))]
    charge_index = [False for t in range(n_s)]
    for k in range(len(include)):    
        i = include[k]
        states = gen[i]['states'][-1]
        if gen[i]['type'] == 'Utility':
            min_c[k] = [0.5*gen[i]['x']['f']*j for j in scale[i]]
            max_c[k] = [1.5*gen[i]['x']['f']*j for j in scale[i]]
        else:
            min_c[k] = [gen[i][states[0]]['f'][-1]*j for j in scale[i]]
            l_state = gen[i][states[-1]]
            max_c[k] = [(l_state['f'][-1]+l_state['ub'][-1]*l_state['h'][-1])*j for j in scale[i]]
        if all([j==0 for j in max_c[k]]):
            f = list(gen[i]['output'].keys())
            for out in f:
                conv = [-j for j in gen[i]['output'][out][-1]]
                if all([j>=0 for j in conv]):
                    min_c[k] = [marginal[out]['min']*min(conv) for t in range(n_s)]
                    max_c[k] = [marginal[out]['max']*max(conv) for t in range(n_s)]
        if any([t<0 for t in min_c[k]]):
            j = 0
            while j<len(states) and any([t<0 for t in min_c[k]]):
                j = j+1
                min_c[k] = [gen[i][states[j]]['f'][-1]*st for st in scale[i]]
            for t in range(len(min_c[k])):
                if min_c[k][t]<0:
                    min_c[k][t] = 1e-3
    if not dispatch is None:
        for i in stor_include:
            for t in range(n_s):
                if (dispatch[i][t+1]-dispatch[i][t])>0:
                    charge_index[t] = True
        
        for k in range(len(include)):
            i = include[k]
            for t in range(n_s):
                if charge_index[t] and dispatch[i][t+1]>0:
                    max_c[k][t] = 1.25*max_c[k][t]

    if not dispatch is None and any([any([dispatch[k][j+1]>0 for j in range(n_s)]) for k in include]):
        unnec = True
    else:
        unnec = False
        
    min_on_t = []
    dt_on = []
    max_on_t = []
    for t in range(n_s):
        if unnec and any([dispatch[k][t+1]>0 for k in include]):
            min_on_t.append(min([min_c[k][t] for k in range(len(include)) if dispatch[include[k]][t+1]>0]))
            max_on_t.append(max([max_c[k][t] for k in range(len(include)) if dispatch[include[k]][t+1]>0]))
            dt_on.append(dt[t])
        elif not unnec:
            if len(include)==0:
                min_on_t.append(0)
                max_on_t.append(0)
            else:
                min_on_t.append(min([min_c[k][t] for k in range(len(include))]))
                max_on_t.append(max([max_c[k][t] for k in range(len(include))]))
            dt_on.append(dt[t])
    m_c = {}
    m_c['min'] = sum([min_on_t[i]*dt_on[i] for i in range(len(dt_on))])/sum(dt_on)
    m_c['max'] = sum([max_on_t[i]*dt_on[i] for i in range(len(dt_on))])/sum(dt_on)

    ac_dc = None
    for i in range(n_g):
        if gen[i]['type']== 'ACDCConverter':
            ac_dc = gen[i]
            break
    for i in range(n_g):
        if gen[i]['type']== 'Utility':
            sc = None
            if s in gen[i]['output']:
                sc = scale[i]
            elif not ac_dc is None and s == 'e' and 'dc' in gen[i]['output']:
                sc = [j*ac_dc['output']['e'][0][1] for j in scale[i]]
            elif not ac_dc is None and s == 'dc' and 'e' in gen[i]['output']:
                sc = [j*ac_dc['output']['dc'][0][0] for j in scale[i]]
            if not sc is None:
                if len(include)==0:
                    m_c = bound_by_utility(None,gen[i],sc)
                else:
                    m_c = bound_by_utility(m_c,gen[i],sc)
    return m_c

def bound_by_utility(m_c,gen,scale):
    if m_c is None:
        m_c = {}
        m_c['max'] = 1.5*gen['x']['f']*max(scale)
        if 'y' in gen:
            m_c['min'] = -gen['y']['f']*min(scale)
        else:
            m_c['min'] = 0.5*gen['x']['f']*min(scale)
    else:
        m_c['max'] = min([m_c['max'], gen['x']['f']*max(scale)])
        if 'y' in gen:
            m_c['min'] = max([m_c['min'],-gen['y']['f']*min(scale)])
        else:
            m_c['min'] = min([m_c['min'], gen['x']['f']*min(scale)])
    return m_c


"""Result template logic.
"""

def result_template(all_data_nodes,names, zones, pl):
    """Returns an empty dictionary with the necessary result fields."""
    weather_headers = ['glo_horz_irr','dir_norm_irr','dif_horz_irr','t_dryb','t_dewp','rh','pres_pa','wdir','wspd','tot_cld','opq_cld']
    f_build = ['avg_T', 'electric', 'cooling', 'heating', 'Tzone', 'return_', 'supply']
    res = dict(
        timestamp = [],
        nodedata = dict((k, {}) for k in all_data_nodes),
        weather = dict((k, []) for k in weather_headers),
        building= dict((k,dict((k,[]) for k in f_build)) for k in names['buildings']),
        fluid_loop = dict((fl,[]) for fl in names['fluid_loop']),
        generator_state = dict((g,[]) for g in names['components']),
        storage_state = dict((n,[]) for n in names['storage']),
        hydro_spill = dict((h,[]) for h in names['hydro']),
        hydro_outflow = dict((h,[]) for h in names['hydro']),
        line_flow = dict((l,[]) for l in names['lines']),
        line_loss = dict((l,[]) for l in names['lines']),
    )
    for i,n in enumerate(names['buildings']):
        res['building'][n]['Tzone'] = [[] for j in range(zones[i])]
        res['building'][n]['return_'] = [[] for j in range(pl[i])]
        res['building'][n]['supply'] = [[] for j in range(pl[i])]
    return res




import urllib.request
import time

from eagers.config.path_spec import USER_DIR_DATA_RETRIEVAL, USER_DIR_HYDRO_DATA


def query_web(html_list,file_list):
    for i in range(len(html_list)):
        urllib.request.urlretrieve(html_list[i],file_list[i])
        time.sleep(1)

def build_lists(sites,param,save_dir,date_range):
    #build list of URL's and file names
    base = 'https://nwis.waterdata.usgs.gov/usa/nwis/uv/?'
    dr = '&period=&begin_date=' + date_range[0] +'&end_date=' + date_range[1]
    html_list = []
    file_list = []
    for k in range(len(sites['names'])):
        if not sites['upriver'][k] is None:
            url = base + 'cb_' + '00060' + '=on&' + 'format=rdb&site_no=' + str(sites['upriver'][k]) + dr
            f_name = save_dir + '/' + sites['names'][k] + param + 'upriver' + '.txt'
            html_list.append(url)
            file_list.append(f_name)
        if not sites['downriver'][k] is None:
            url = base + 'cb_' + '00060' + '=on&' + 'format=rdb&site_no=' + str(sites['downriver'][k]) +dr
            f_name = save_dir + '/' + sites['names'][k] + param + 'downriver' + '.txt'
            html_list.append(url)
            file_list.append(f_name)
        if not sites['reservoir'][k] is None:
            url = base + 'cb_' + '00062' + '=on&' + 'format=rdb&site_no=' + str(sites['reservoir'][k]) +dr
            f_name = save_dir + '/' + sites['names'][k] + param + 'reservoir' + '.txt'
            html_list.append(url)
            file_list.append(f_name)
    return html_list, file_list

def site_names(n):
    sites = {}
    sites['names'] = ['Lake_Chelan_Dam',
                    'Cabinet_Gorge_Dam',
                    'Noxon_Rapids_Dam',
                    'Thompson_Falls_Dam',
                    'Bonneville_Dam',
                    'Chief_Joseph_Dam',
                    'Grand_Coulee_Dam',
                    'John_Day_Dam',
                    'Keenleyside_Dam',
                    'McNary_Dam',
                    'Mica_Dam',
                    'Priest_Rapids_Dam',
                    'Revelstoke_Dam',
                    'Rock_Island_Dam',
                    'Rocky_Reach_Dam',
                    'The_Dalles_Dam',
                    'Wanapum_Dam',
                    'Wells_Dam',
                    'Pelton_Dam',
                    'Hungry_Horse_Dam',
                    'Kerr_Dam',
                    'Brilliant_Dam',
                    'Corra_Linn_Dam',
                    'Kootenay_Canal_Generating_Station_Dam',
                    'Libby_Dam',
                    'South_Slocan_Dam',
                    'Upper_Bonnington_Falls_Dam',
                    'Dworshak_Dam',
                    'Long_Lake_Dam',
                    'Owyhee_Dam',
                    'Albeni_Falls_Dam',
                    'Boundary_Dam',
                    'Box_Canyon_Dam',
                    'Seven_Mile_Dam',
                    'Waneta_Dam',
                    'American_Falls_Dam',
                    'Bliss_Dam',
                    'Brownlee_Dam',
                    'C._J._Strike_Dam',
                    'Hells_Canyon_Dam',
                    'Ice_Harbor_Dam',
                    'Little_Goose_Dam',
                    'Lower_Granite_Dam',
                    'Lower_Monumental_Dam',
                    'Lower_Salmon_Falls_Dam',
                    'Milner_Dam',
                    'Oxbow_Dam',
                    'Shoshone_Falls_Dam',
                    'Twin_Falls_Dam']
    sites['upriver'] = [12451000,
                        None,
                        None,
                        
                        
                        ]

    sites['downriver'] = [12452500,
                          12391950,
                          12391400,

                        ]

    sites['reservoir'] = [12452000,
                          None,
                          None,
                          ]

    active_sites = {}
    for f in sites:
        active_sites[f] = [sites[f][j] for j in range(n)]
    return active_sites

from openpyxl import load_workbook


save_dir = USER_DIR_DATA_RETRIEVAL

usgs_param = {"discharge": '00060',"guage_height":'00065',"reservoir_elevation":'00062'}
wb = load_workbook(str(USER_DIR_HYDRO_DATA / 'usgs.xlsx'))
sites = site_names(2)
date_range = ['2007-01-01','2020-10-08']
html_list,file_list = build_lists(sites,usgs_param['discharge'],save_dir,date_range)
#query website
query_web(html_list,file_list)
#parse files
''' If dam A has an upstream gauge and the upstream dam has a downstream gauge, 
then source/sink can be calculated at the time resolution of the data (accounting 
for river run time). Otherwise if run of river average out daily difference in 
outflow to get avg daily source/sink'''
    #organize to common timestep and horizon



from eagers.building.process.load_sched import load_sched

def initial_param(building,T_zone1,T_surf1):
    '''
    initial conditions for some more state variables
    '''
    T ={}
    air_nodes ={}
    plant_nodes={}
    nl = len(building.plant_loop['name'])
    T_supply = [[] for i in range(nl)]
    T_return = [[] for i in range(nl)]
    T['zone'] = T_zone1
    T['surf'] = T_surf1
    n_win = len(building.windows['name'])
    T['windows'] = [[300,300,300,300] for i in range(n_win)]
    T['windows_int'] = [300 for i in range(n_win)]
    for i in range(n_win):
        if building.windows['type'][i]=='SimpleGlazingSystem':
            T['windows'][i][2:] = [0,0]
    
    sn = len(building.plant_supply_nodes['name'])
    dn = len(building.plant_demand_nodes['name'])
    plant_nodes['demand_temperature'] = [0 for i in range(dn)]
    plant_nodes['demand_flow'] = [0 for i in range(dn)]
    plant_nodes['supply_temperature'] = [0 for i in range(sn)]
    for i in range(len(building.plant_loop['name'])):
        T_supply[i] = building.plant_loop['exit_temperature'][i]
        if building.plant_loop['type'][i].lower()=='heating': #heating loop, water comes back cooler
            T_return[i] = building.plant_loop['exit_temperature'][i]-building.plant_loop['temperature_difference'][i]
        else:
            T_return[i] = building.plant_loop['exit_temperature'][i]+building.plant_loop['temperature_difference'][i]
    air_nodes['supply_T_set'] = [0 for i in range(sn)]
    air_nodes['supply_w_set'] = [0 for i in range(sn)]
    air_nodes['supply_flow'] = [0 for i in range(sn)]
    air_nodes['demand_T_set'] = [0 for i in range(dn)]
    air_nodes['demand_w_set'] = [0 for i in range(dn)]
    air_nodes['demand_flow'] = [0 for i in range(dn)]
    return T,air_nodes,plant_nodes,T_supply,T_return



'''
Run a simulation with forecasting and actual data
'''
from eagers.read.get_data import get_data
from eagers.building.process.load_sched import load_sched
from eagers.building.process.initial_param import initial_param
from eagers.building.process.reset_loop_param import reset_loop_param
from eagers.building.process.zone_loads import zone_loads
from eagers.building.process.forecast_dr_capacity import forecast_dr_capacity
from eagers.building.process.buildings_step import buildings_step
from eagers.building.process.load_actual_building_data import update_real_gains

import datetime

def run_sim(building,test_data,weather,T_zone,T_surf,humidity,date,horizon):
    facility = {}
    e_use = {}
    weather_now = {}
    weather_forecast = {}
    frost = [[0 for i in range(len(building.cases))]]
    net_elec = []
    heat_absorbed_by_loop = []
    cooling_absorbed_by_loop = []
    w = {}
    w['zone'] = humidity[0]
    f_cat = ['net_elec','heat_elec','heat_gas','cool_elec','fan_elec','other_fans','water_gas','water_elec','tower_elec','pumps','heating_load','cooling_load','tower_load','supply_T','return_T']
    for i in f_cat:
        facility[i] = []
    T,air_nodes,plant_nodes,T_supply,T_return = initial_param(building,T_zone[0],T_surf[0])
    facility['supply_T'].append(T_supply)
    facility['return_T'].append(T_return)
    t_last = date[0]
    for t in range(len(date)-1):#number of simulation steps
        dt = (date[t+1] - t_last).seconds# resolution in seconds
        d_now = date[t+1] 
        d_forecast = [d_now + datetime.timedelta(hours=i) for i in range(horizon+1)]
        for i in weather:
            weather_now[i] = weather[i][t]
            weather_forecast[i] = [weather[i][t+j] for j in range(horizon+1)]

        observer = {}
        observer['building_humidity'][0] = w['zone']
        observer['building_zone_temp'][0] = T_zone[t]
        observer['building_surf_temp'][0] = T_surf[t]
        observer['building_supply'][0] = T_supply[t]
        forecast,T_building_effective,heating_t,cooling_t = forecast_dr_capacity([building],observer,weather_forecast,d_forecast)
        Tz_nominal = forecast['Tz_nominal']
        ## compute vertices

        ## receive heating/cooling and 'optimal' building temperature values from transactive optimization
        n0 = int((len(heating_t[0][0])-1)/2)
        heating = heating_t[0][0][n0]
        cooling = cooling_t[0][0][n0]
        T_optimal = T_building_effective[0][0][n0]
        
        ##Update actual data
        actual_data = get_data(test_data,[d_now])[0]
        T,w,air_nodes,plant_nodes = reset_loop_param(building,T,w,air_nodes,plant_nodes,weather_now,d_now)
        schedules = load_sched(building.schedule,building.holidays,[d_now],None,False)
        loads,gains,_,_,_,_,_,_,_ = zone_loads(building,d_now,schedules,actual_data['weather'],T_zone[t],humidity[t],dt,frost[t])
        new_loads,new_gains,_ = update_real_gains(building,actual_data['weather'],loads,gains,d_now,actual_data['internal_gains'])
        ##simulate building forward in time
        T, humid_t1, frost_t, e_use, loop_load, air_nodes, plant_nodes = buildings_step(building,actual_data['weather'],T,w,air_nodes,plant_nodes,e_use,new_loads,new_gains,heating,cooling,T_optimal,Tz_nominal,schedules,frost[t],dt,d_now)

        ##Collect and organize
        frost.append(frost_t)
        T_zone.append(T['zone'])
        T_surf.append(T['surf'])
        humidity.append(humid_t1)
        for i in range(10):
            facility[f_cat[i]].append(e_use[f_cat[i]])
        facility['heating_load'].append(sum(loop_load[0]))
        facility['cooling_load'].append(sum(loop_load[1]))
        facility['tower_load'].append(sum(loop_load[2]))
        T_supply = []
        T_return = []
        for l in range(len(building.plant_loop['name'])):
            T_return.append(plant_nodes['supply_temperature'][building.plant_supply_nodes['inlet_node'][l]])
            T_supply.append(plant_nodes['demand_temperature'][building.plant_demand_nodes['inlet_node'][l]])
        facility['supply_T'].append(T_supply)
        facility['return_T'].append(T_return)
        net_elec.append((facility['net_elec'][t] - facility['cool_elec'][t])/1000) #kW
        heat_absorbed_by_loop.append(heating - facility['heating_load'][t]/1000)
        cooling_absorbed_by_loop.append(cooling - facility['cooling_load'][t]/1000)
        t_last = date[t+1]
    return facility, net_elec


"""Contains structure descriptions for results to be saved in HDF5
format.
"""

import tables as tb


# ----------------
# Module constants
# ----------------

# Conversion from shape codes to dimension info field names.
# These provide the mapping between given dimension info field names and
# the letters used in column shape specifications.
#
# ASSUMPTION:
# column_realization assumes that every shape code is a single letter.
# Be careful when adding new ones to either follow the current
# convention or update it.
SHAPECODE_TO_DIMFIELD = dict(
    h='horizon',
    z='zone',
    p='plantloop', #largest number of plant water loops at any building
    t='timer',
)


# ------------------------------------------
# Table column description factory functions
# ------------------------------------------
#
# If a column is specified as a dictionary, shape can be specified as a
# tuple of strings.  The SHAPECODE_TO_DIMFIELD variable defines the
# mapping between field names specified in the given dimension
# information dictionary and letters used in column shape
# specifications.
#
# ASSUMPTIONS:
# column_realization assumes:
#   - every dimension specification contains at most one shape code.
#       E.g. shape=('g+h',) will not work.
# Be careful when modifying column descriptions to either follow the
# current convention or update it.

def nodedata_descr(node_names,v_shape,v_pos):
    def make_nd_dict(pos,v_shape):
        f_list = ['demand','demand_history','inflow','outflow','inflow_history','outflow_history']
        desc = dict(_v_pos=pos,)
        for j,f in enumerate(f_list):
            desc[f] = dict(col='Float64Col', pos=j)
            if not v_shape is None:
                desc[f]['shape'] = (v_shape)
        return desc
    nddesc = {}
    for i, n in enumerate(node_names):
        nddesc[n] = make_nd_dict(i,v_shape)
    nddesc['_v_pos'] = v_pos
    return nddesc


def weather_descr(v_shape,v_pos):
    weather = {}
    w_names = ['glo_horz_irr', 'dir_norm_irr','dif_horz_irr','t_dryb','t_dewp','rh','pres_pa','wdir','wspd','tot_cld','opq_cld']
    for k,n in enumerate(w_names):
        weather[n] = dict(col='Float32Col', pos=k)
        if not v_shape == None:
            weather[n]['shape'] = (v_shape)
    weather['_v_pos'] = v_pos
    return weather

def gen_descr(names,v_shape,v_pos):
    gendesc = {}
    for k,n in enumerate(names):
        gendesc[n] = dict(col='Float64Col', pos=k)
        if not v_shape == None:
            gendesc[n]['shape'] = (v_shape)
    gendesc['_v_pos'] = v_pos
    return gendesc

def building_descr(build_names, z_dif, pl_dif,v_shape,v_pos):
    def make_build_dict(pos,v_shape,z_dif,pl_dif):
        if v_shape is None:
            f_list = ['avg_T','Tzone','electric','cooling','heating','return_','supply']
            s_list = [None,'z',None,None,None,'p','p']
        else:
            if z_dif is None:
                f_list = ['heating','cooling','temperature']
                s_list = [None,None,None]
            else:
                f_list = ['avg_T','Tzone','electric','cooling','heating']
                s_list = [None,'z',None,None,None]
        
        desc = dict(_v_pos=pos,)
        for j,f in enumerate(f_list):
            if s_list[j] is None:
                desc[f] = dict(col='Float64Col', pos=j)    
                if not v_shape is None:
                    desc[f]['shape'] = (v_shape)
            else:
                sh = s_list[j]
                if sh == 'z' and z_dif>0:
                    sh = sh + '-' + str(int(z_dif))
                elif sh == 'p' and pl_dif>0:
                    sh = sh + '-' + str(int(pl_dif))
                if v_shape is None:
                    desc[f] = dict(col='Float64Col', pos=j, shape=(sh))
                else:
                    desc[f] = dict(col='Float64Col', pos=j, shape=[v_shape,sh])
        return desc
    build = {}
    for i, n in enumerate(build_names):
        if z_dif is None:
            build[n] = make_build_dict(i,v_shape,None,None)
        else:
            build[n] = make_build_dict(i,v_shape,z_dif[i],pl_dif[i])
    build['_v_pos'] = v_pos
    return build


def dispatch_descr(names, z_dif, pl_dif):
    dispatch = dict(timestamp=tb.Time64Col(pos=0))
    dispatch['generator_state'] = gen_descr(names['components'],None,1)
    dispatch['storage_state'] = gen_descr(names['storage'],None,2)
    dispatch['line_flow'] = gen_descr(names['lines'],None,3) 
    dispatch['fluid_loop'] = gen_descr(names['fluid_loop'],None,4) 
    dispatch['building'] = building_descr(names['buildings'], z_dif, pl_dif,None,5)
    dispatch['nodedata'] = nodedata_descr(names['nodes'],None,6)
    dispatch['weather']= weather_descr(None,7)
    return dispatch


def predicted_descr(names, z_dif, pl_dif,h):
    predicted = dict(timestamp = tb.Time64Col(pos=0, shape=h))
    predicted['cost'] = dict(col='Time64Col', pos=1,)
    predicted['lb_relax'] = dict(col='Time64Col', pos=2,)
    predicted['generator_state'] = gen_descr(names['components'],'h',3)
    predicted['storage_state'] = gen_descr(names['storage'],'h',4)
    predicted['line_flow'] = gen_descr(names['lines'],'h',5) 
    predicted['building'] = building_descr(names['buildings'], z_dif, pl_dif,'h',6)
    predicted['fluid_loop'] = gen_descr(names['fluid_loop'],'h',7) 
    predicted['nodedata'] = nodedata_descr(names['nodes'],'h',8)
    predicted['weather'] = weather_descr('h',9)
    return predicted


def solution_descr(names,h):
    sol = dict(timestamp=tb.Time64Col(pos=0, shape=h+1),
               timer=tb.Time64Col(pos=0, shape=3))
    sol['generator_state'] = gen_descr(names['components'],'h',2)
    sol['storage_state'] = gen_descr(names['storage'],'h',3)
    sol['excess_heat'] = gen_descr(names['heating_nodes'],'h',4) 
    sol['excess_cool'] = gen_descr(names['cooling_nodes'],'h',5) 
    sol['line_flow'] = gen_descr(names['lines'],'h',6) 
    sol['line_loss'] = gen_descr(names['lines'],'h',7) 
    sol['fluid_loop'] = gen_descr(names['fluid_loop'],'h',8) 
    sol['building'] = building_descr(names['buildings'], None, None,'h',9)
    return  sol


# ----------------
# Module functions
# ----------------

def result_description(names, dimensions, zones, pl):
    """Return the result description, accounting for the different
    project-related dimensions specified.

    Positional arguments:
    node_names - (list of str) Node names.
    dimensions - (dict) Dictionary specifying dimension counts for
        different project elements.
    """
    z_dif = [dimensions['zone'] - z for z in zones]
    pl_dif = [dimensions['plantloop'] - n for n in pl]
    h = dimensions['horizon']
    # Update dispatch description.
    disp = realize_all_columns(dispatch_descr(names, z_dif, pl_dif), dimensions)
    # Update predicted description.
    pred = realize_all_columns(predicted_descr(names, z_dif, pl_dif,h), dimensions)
    # Update solution description.
    solu = realize_all_columns(solution_descr(names,h), dimensions)
    # Create and return result description.
    return disp, pred, solu


def realize_all_columns(descr, dimensions):
    """Use dictionary column specifications to create PyTables column
    instances.

    Positional arguments:
    descr - (dict) Description specification whose columns are to be
        realized.
    dimensions - (dict) Dimensional count information.
    """
    realized = {}
    for k, v in descr.items():
        if isinstance(v, dict):
            if 'col' in v:
                col_realzn = column_realization(v, dimensions)
            else:
                col_realzn = realize_all_columns(v, dimensions)
                if set(col_realzn.keys()) == {'_v_pos'}:
                    # Only key is '_v_pos'.
                    del col_realzn['_v_pos']
            if col_realzn:
                realized[k] = col_realzn
        else:
            realized[k] = v
    return realized


def column_realization(col, dimensions):
    """Return a PyTables column instance based on the given
    specification.

    Assumptions:
    - Every shape code is a single letter.
    - Every column shape specification contains at most one shape code,
        which comes at the beginning of each dimension specification.

    Positional arguments:
    col - (dict) Column specification. This will be converted to a
        PyTables column instance.
    dimensions - (dict) Dimensional information for this column.
    """
    coltype = getattr(tb, col.pop('col'))
    try:
        shape_spec = col['shape']
    except KeyError:
        return coltype(**col)
    # Create shape based on given dimensional information.
    new_shape = []
    if isinstance(shape_spec,(str,list)):
        for spec in shape_spec:
            # Evaluate the shape specification.
            shapecode = spec[0]
            dimfield = SHAPECODE_TO_DIMFIELD.get(shapecode)
            if dimfield:
                # Shape code corresponds to a potential dimension key.
                dim = dimensions.get(dimfield)
                if dim:
                    # A dimension number was given for the key.
                    new_spec = f"{dim}{spec[1:]}"
                else:
                    return
            else:
                new_spec = spec
            new_shape.append(eval(new_spec))
    if any(new_shape):
        col['shape'] = tuple(new_shape)
        return coltype(**col)
    # If all dimensions are zero, None is returned.



import numpy as np
from scipy.linalg import expm, solve


def ss_response(size_kw, lower_bound, ramp_rate, a_, b_, c, d):
    """Calculate state space response for a given Component.

    Positional arguments:
    gen - (Component) Component (used to be generator) for which the
        response is to be calculated.
    """
    # Get timestep.
    rr0 = ramp_rate+0
    t_peak = (size_kw - lower_bound) / ramp_rate * 3600
    if t_peak > 4*60:
        dt = 60
    elif t_peak > 3.6e4:
        dt = 3600
    else:
        dt = 1
    # Convert continuous state-space to discrete time state-space.
    a = expm(np.array(a_) * dt).tolist()
    b = (solve(np.array(a_), (a - np.eye(2)))
        @ np.array(b_)).tolist()

    time = 5 * t_peak
    ramp_rate = False
    count = 0
    while isinstance(ramp_rate, bool):
        n_s = int(round(time / dt, 0))
        u = [size_kw] * (n_s + 1)
        x_0 = []
        for i in range(len(c)):
            if c[i] == 1:
                x_0.append(lower_bound)
            else:
                x_0.append(0)
        y, t = ss_sim(a, b, [c], [d], x_0, u, dt)
        norm_end = [abs(y[j][0]-y[-1][0])/y[-1][0] for j in range(len(u)-10,len(u))]
        if (y[-1][0] - y[0][0])/(u[0] - y[0][0]) > 0.95 and max(norm_end)<1e-3:
            n_r = 0
            n_val = 0
            while n_val < 0.95:
                n_r+=1
                n_val = (y[n_r][0] - y[0][0])/(u[0] - y[0][0])
            r = 1-(n_val - .95)/(n_val - (y[n_r-1][0] - y[0][0])/(u[0] - y[0][0]))
            t_rise = (t[n_r-1] + r*(t[n_r] - t[n_r-1]))/3600 #rise time in hours
            ramp_rate = (size_kw - lower_bound)*0.95/t_rise

        else:
            time *= 5
        count += 1
        if count > 10:
            ramp_rate = rr0
    return ramp_rate


def ss_sim(a, b, c, d, x_0, u, dt):
    time = [t*dt for t in range(len(u)+1)]
    x = [[0 for j in range(len(x_0))] for i in range(len(u) + 1)]
    x[0] = x_0
    y = [[0 for j in range(len(c))] for t in range(len(u) + 1)]
    y[0] = [sum([c[j][i] * x_0[i] for i in range(len(x_0))]) for j in range(len(c))]
    for t in range(len(u)):
        x[t+1] = [sum([a[i][j] * x[t][j] for j in range(len(x_0))]) + b[i] * u[t] for i in range(len(a))] 
        y[t+1] = [sum([c[j][i] * x[t+1][i] for i in range(len(c))]) + d[j] * u[t] for j in range(len(c))]
    return y, time



def wind_speed_calc(wind,height,terrain):
    d_met = 270 #boundary layer thickness at meterological site
    z_met = 10 #elevation of met site
    a_met = 0.14 #coefficient at met site
    if terrain =='city':
            a = .33
            d = 460
    elif terrain =='urban':
            a = .22
            d = 370
    elif terrain =='ocean':
            a = 0.1
            d = 210
    elif terrain =='flat':
            a = 0.14
            d = 270
    elif terrain =='rough':
            a = 0.22
            d = 370
    else:
         print('what is terrain?')
    wind_speed = [wind*(d_met/z_met)**a_met*(max(0,h)/d)**a for h in height]
    return wind_speed


"""Defines network-related constants."""


# Network names, their corresponding abbreviation, what they
# represent:
# Electrical --- E  --- standard 480V AC electrical network
# DirectCurrent --- DC --- 48V DC electrical network
# Transmission1 --- E1 --- 230kV electric transmission (E2, E3, etc
#                             can be additional voltage levels)
# DistrictHeat --- H --- standard 80C supply heating bus
# DistrictCool --- C --- standard 4C supply cooling bus
# Heating2     --- H2 --- Heat a different temperature than
#                         DistrictHeat (H3, H4... as needed)
# CoolingWater --- CW --- Water circulated between chillers and
#                         cooling towers
# Hydro        --- W --- River network with reservoirs and dams
# Hydrogen     --- Hy --- Gaseous hydrogen stream
# LiqHydrogen  --- LH2 --- Liquid hydrogen
NETWORK_NAMES = (
    'electrical', 'direct_current', 'transmission_1',
    'district_heat', 'district_cool', 'heating_2', 'cooling_water',
    'hydro',
    'hydrogen', 'liq_hydrogen',
)
NETWORK_ABBRS = (
    'e', 'dc', 'e1',
    'h', 'c', 'h2', 'cw',
    'w',
    'hy', 'lh2',
)
NETWORK_ABBR_NAME_MAP = dict(zip(NETWORK_ABBRS, NETWORK_NAMES))
NETWORK_NAME_ABBR_MAP = dict(zip(NETWORK_NAMES, NETWORK_ABBRS))

# Mapping between Component outputs and the networks they would export
# to.
OUTPUT_NAMES = (
    'electricity', 'direct_current', None,
    'heat', 'cooling', None, None,
    'hydro',
    'hydrogen', 'liq_hydrogen',
)
OUTPUT_NETWORK_MAP = dict(zip(OUTPUT_NAMES, NETWORK_NAMES))


"""Defines Building class.
"""

from eagers.class_definition.data_constructed import DataConstructed
from eagers.read.import_idf import import_idf
from eagers.read.import_sizes import import_sizes
from eagers.basic.file_handling import ensure_suffix


class Building(DataConstructed):
    def __init__(self,
            *args,
            name=None,
            site=None,
            sim_date=None,
            location=None,
            convection=None,
            schedule=None,
            holidays=None,
            material=None,
            window_material=None,
            surfaces=None,
            windows=None,
            doors=None,
            window_frames=None,
            zones=None,
            infiltration=None,
            mixing=None,
            occupancy=None,
            lighting_internal=None,
            plug_load=None,
            gas_load=None,
            exterior=None,
            cases=None,
            walk_ins=None,
            racks=None,
            water_use=None,
            hvac=None,
            air_demand_nodes=None,
            air_demand_equip=None,
            air_supply_nodes=None,
            air_supply_equip=None,
            plant_loop=None,
            plant_demand_nodes=None,
            plant_demand_equip=None,
            plant_supply_nodes=None,
            plant_supply_equip=None,
            pump=None,
            chiller=None,
            boiler=None,
            humidifiers=None,
            unitary_heat_cool=None,
            cooling_tower=None,
            water_heater=None,
            controller=None,
            manager=None,
            zone_controls=None,
            thermostat=None,
            setpoints=None,
            unitary_sys=None,
            fans=None,
            curves=None,
            coils_cooling=None,
            coils_heating=None,
            ground_temperatures=None,
            water_main_temperature=None,
            impact_factor=None,
            ctf=None,
            **kwargs,
            ):
        super().__init__(*args, **kwargs)
        self.name = name
        self.site = site
        self.sim_date = sim_date
        self.location = location
        self.convection = convection
        self.schedule = schedule
        self.holidays = holidays
        self.material = material
        self.window_material = window_material
        self.surfaces = surfaces
        self.windows = windows
        self.doors = doors
        self.window_frames = window_frames
        self.zones = zones
        self.infiltration = infiltration
        self.mixing = mixing
        self.occupancy = occupancy
        self.lighting_internal = lighting_internal
        self.plug_load = plug_load
        self.gas_load = gas_load
        self.exterior = exterior
        self.cases = cases
        self.walk_ins = walk_ins
        self.racks = racks
        self.water_use = water_use
        self.hvac = hvac
        self.air_demand_nodes = air_demand_nodes
        self.air_demand_equip = air_demand_equip
        self.air_supply_nodes = air_supply_nodes
        self.air_supply_equip = air_supply_equip
        self.plant_loop = plant_loop
        self.plant_demand_nodes = plant_demand_nodes
        self.plant_demand_equip = plant_demand_equip
        self.plant_supply_nodes = plant_supply_nodes
        self.plant_supply_equip = plant_supply_equip
        self.pump = pump
        self.chiller = chiller
        self.boiler = boiler
        self.humidifiers = humidifiers
        self.unitary_heat_cool = unitary_heat_cool
        self.cooling_tower = cooling_tower
        self.water_heater = water_heater
        self.controller = controller
        self.manager = manager
        self.zone_controls = zone_controls
        self.thermostat = thermostat
        self.setpoints = setpoints
        self.unitary_sys = unitary_sys
        self.fans = fans
        self.curves = curves
        self.coils_cooling = coils_cooling
        self.coils_heating = coils_heating
        self.ground_temperatures = ground_temperatures
        self.water_main_temperature = water_main_temperature
        self.impact_factor = impact_factor
        self.ctf = ctf

    @classmethod
    def from_file(cls, filename):
        """Returns a Building object using data read from IDF and EIO
        files with the same name.
        """
        building = cls(**import_idf(ensure_suffix(filename, '.idf')))
        return import_sizes(ensure_suffix(filename, '.eio'), building)

    @classmethod
    def from_energyplus(cls, filename):
        """More obvious name for from_file() method."""
        return cls.from_file(filename)
